@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2018-2021 TileDB, Inc.
+ * @copyright Copyright (c) 2018-2022 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -34,6 +34,10 @@
 #include <catch.hpp>
 #include "test/src/helpers.h"
 #include "tiledb/sm/filesystem/vfs.h"
+#ifdef _WIN32
+#include "tiledb/sm/filesystem/path_win.h"
+#endif
+#include "tiledb/sm/tile/tile.h"
 
 using namespace tiledb::common;
 using namespace tiledb::sm;
@@ -52,29 +56,30 @@ TEST_CASE("VFS: Test read batching", "[vfs]") {
 
   bool exists = false;
   REQUIRE(vfs->is_file(testfile, &exists).ok());
-  if (exists)
+  if (exists) {
     vfs->remove_file(testfile);
+  }
 
   // Write some data.
   const unsigned nelts = 100;
-  uint32_t data_write[nelts], data_read[nelts];
-  for (unsigned i = 0; i < nelts; i++)
+  uint32_t data_write[nelts];
+  for (unsigned i = 0; i < nelts; i++) {
     data_write[i] = i;
+  }
   REQUIRE(vfs->write(testfile, data_write, nelts * sizeof(uint32_t)).ok());
   REQUIRE(vfs->terminate().ok());
 
   Tile tile[nelts];
   for (uint64_t i = 0; i < nelts; i++) {
-    Buffer buff(&data_read[i], nelts - i);
-    tile[i].filtered_buffer()->swap(buff);
+    tile[i].filtered_buffer().expand(nelts * sizeof(uint32_t));
   }
 
-  std::vector<std::tuple<uint64_t, Tile*, uint64_t>> batches;
+  std::vector<tuple<uint64_t, Tile*, uint64_t>> batches;
   std::vector<ThreadPool::Task> tasks;
 
   SECTION("- Default config") {
     // Check reading in one batch: single read operation.
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
+    std::memset(tile[0].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
     batches.emplace_back(0, &tile[0], nelts * sizeof(uint32_t));
     REQUIRE(
         vfs->init(&g_helper_stats, &compute_tp, &io_tp, nullptr, nullptr).ok());
@@ -82,11 +87,12 @@ TEST_CASE("VFS: Test read batching", "[vfs]") {
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
     for (unsigned i = 0; i < nelts; i++)
-      REQUIRE(data_read[i] == i);
+      REQUIRE(tile[0].filtered_buffer().data_as<uint32_t>()[i] == i);
 
     // Check reading first and last element: 1 reads due to the default
     // batch size.
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
+    std::memset(tile[0].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
+    std::memset(tile[1].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
     batches.clear();
     batches.emplace_back(0, &tile[0], sizeof(uint32_t));
     batches.emplace_back(
@@ -94,20 +100,23 @@ TEST_CASE("VFS: Test read batching", "[vfs]") {
     REQUIRE(vfs->read_all(testfile, batches, &io_tp, &tasks).ok());
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
-    REQUIRE(data_read[0] == 0);
-    REQUIRE(data_read[1] == nelts - 1);
+    REQUIRE(tile[0].filtered_buffer().data_as<uint32_t>()[0] == 0);
+    REQUIRE(tile[1].filtered_buffer().data_as<uint32_t>()[0] == nelts - 1);
 
     // Check each element as a different region: single read because there is no
     // amplification required (all work is useful).
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
     batches.clear();
-    for (unsigned i = 0; i < nelts; i++)
+    for (unsigned i = 0; i < nelts; i++) {
+      std::memset(
+          tile[i].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
       batches.emplace_back(i * sizeof(uint32_t), &tile[i], sizeof(uint32_t));
+    }
     REQUIRE(vfs->read_all(testfile, batches, &io_tp, &tasks).ok());
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
-    for (unsigned i = 0; i < nelts; i++)
-      REQUIRE(data_read[i] == i);
+    for (unsigned i = 0; i < nelts; i++) {
+      REQUIRE(tile[i].filtered_buffer().data_as<uint32_t>()[0] == i);
+    }
     REQUIRE(vfs->terminate().ok());
   }
 
@@ -125,30 +134,35 @@ TEST_CASE("VFS: Test read batching", "[vfs]") {
                 .ok());
 
     // Check large batches are not split up.
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
+    std::memset(tile[0].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
     batches.emplace_back(0, &tile[0], nelts * sizeof(uint32_t));
     REQUIRE(vfs->read_all(testfile, batches, &io_tp, &tasks).ok());
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
-    for (unsigned i = 0; i < nelts; i++)
-      REQUIRE(data_read[i] == i);
+    for (unsigned i = 0; i < nelts; i++) {
+      REQUIRE(tile[0].filtered_buffer().data_as<uint32_t>()[i] == i);
+    }
 
     // Check each element as a different region (results in several read
     // operations).
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
     batches.clear();
-    for (unsigned i = 0; i < nelts / 2; i++)
+    for (unsigned i = 0; i < nelts / 2; i++) {
+      std::memset(
+          tile[i].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
       batches.emplace_back(
           2 * i * sizeof(uint32_t), &tile[i], sizeof(uint32_t));
+    }
     REQUIRE(vfs->read_all(testfile, batches, &io_tp, &tasks).ok());
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
-    for (unsigned i = 0; i < nelts / 2; i++)
-      REQUIRE(data_read[i] == 2 * i);
+    for (unsigned i = 0; i < nelts / 2; i++) {
+      REQUIRE(tile[i].filtered_buffer().data_as<uint32_t>()[0] == 2 * i);
+    }
 
     // Check reading first and last element (results in 2 reads because the
     // whole region is too big).
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
+    std::memset(tile[0].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
+    std::memset(tile[1].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
     batches.clear();
     batches.emplace_back(0, &tile[0], sizeof(uint32_t));
     batches.emplace_back(
@@ -156,8 +170,8 @@ TEST_CASE("VFS: Test read batching", "[vfs]") {
     REQUIRE(vfs->read_all(testfile, batches, &io_tp, &tasks).ok());
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
-    REQUIRE(data_read[0] == 0);
-    REQUIRE(data_read[1] == nelts - 1);
+    REQUIRE(tile[0].filtered_buffer().data_as<uint32_t>()[0] == 0);
+    REQUIRE(tile[1].filtered_buffer().data_as<uint32_t>()[0] == nelts - 1);
     REQUIRE(vfs->terminate().ok());
   }
 
@@ -174,15 +188,18 @@ TEST_CASE("VFS: Test read batching", "[vfs]") {
                 .ok());
 
     // There should be a single read due to the gap
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
     batches.clear();
-    for (unsigned i = 0; i < nelts; i++)
+    for (unsigned i = 0; i < nelts; i++) {
+      std::memset(
+          tile[i].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
       batches.emplace_back(i * sizeof(uint32_t), &tile[i], sizeof(uint32_t));
+    }
     REQUIRE(vfs->read_all(testfile, batches, &io_tp, &tasks).ok());
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
-    for (unsigned i = 0; i < nelts; i++)
-      REQUIRE(data_read[i] == i);
+    for (unsigned i = 0; i < nelts; i++) {
+      REQUIRE(tile[i].filtered_buffer().data_as<uint32_t>()[0] == i);
+    }
     REQUIRE(vfs->terminate().ok());
   }
 
@@ -199,15 +216,18 @@ TEST_CASE("VFS: Test read batching", "[vfs]") {
                 .ok());
 
     // There should be a single read due to the batch size
-    std::memset(data_read, 0, nelts * sizeof(uint32_t));
     batches.clear();
-    for (unsigned i = 0; i < nelts; i++)
+    for (unsigned i = 0; i < nelts; i++) {
+      std::memset(
+          tile[i].filtered_buffer().data(), 0, nelts * sizeof(uint32_t));
       batches.emplace_back(i * sizeof(uint32_t), &tile[i], sizeof(uint32_t));
+    }
     REQUIRE(vfs->read_all(testfile, batches, &io_tp, &tasks).ok());
     REQUIRE(io_tp.wait_all(tasks).ok());
     tasks.clear();
-    for (unsigned i = 0; i < nelts; i++)
-      REQUIRE(data_read[i] == i);
+    for (unsigned i = 0; i < nelts; i++) {
+      REQUIRE(tile[i].filtered_buffer().data_as<uint32_t>()[0] == i);
+    }
     REQUIRE(vfs->terminate().ok());
   }
 
@@ -482,3 +502,72 @@ TEST_CASE("VFS: URI semantics", "[vfs][uri]") {
     vfs.terminate();
   }
 }
+
+TEST_CASE("VFS: test ls_with_sizes", "[vfs][ls-with-sizes]") {
+  ThreadPool compute_tp;
+  ThreadPool io_tp;
+  REQUIRE(compute_tp.init(4).ok());
+  REQUIRE(io_tp.init(4).ok());
+  VFS vfs;
+  REQUIRE(
+      vfs.init(&g_helper_stats, &compute_tp, &io_tp, nullptr, nullptr).ok());
+
+#ifdef _WIN32
+  std::string path = tiledb::sm::Win::current_dir() + "\\vfs_test\\";
+#else
+  std::string path =
+      std::string("file://") + tiledb::sm::Posix::current_dir() + "/vfs_test/";
+#endif
+
+  // Clean up
+  bool is_dir;
+  vfs.is_dir(URI(path), &is_dir);
+  if (is_dir)
+    vfs.remove_dir(URI(path));
+
+  std::string dir = path + "ls_dir";
+  std::string file = dir + "/file";
+  std::string subdir = dir + "/subdir";
+  std::string subdir_file = subdir + "/file";
+
+  // Create directories and files
+  vfs.create_dir(URI(path));
+  vfs.create_dir(URI(dir));
+  vfs.create_dir(URI(subdir));
+  vfs.touch(URI(file));
+  vfs.touch(URI(subdir_file));
+
+  // Write to file
+  std::string s1 = "abcdef";
+  REQUIRE(vfs.write(URI(file), s1.data(), s1.size()).ok());
+
+  // Write to subdir file
+  std::string s2 = "abcdef";
+  REQUIRE(vfs.write(URI(subdir_file), s2.data(), s2.size()).ok());
+
+  // List
+  auto&& [status, rv] = vfs.ls_with_sizes(URI(dir));
+  auto children = *rv;
+
+  REQUIRE(status.ok());
+
+#ifdef _WIN32
+  // Normalization only for Windows
+  file = tiledb::sm::path_win::uri_from_path(file);
+  subdir = tiledb::sm::path_win::uri_from_path(subdir);
+#endif
+
+  // Check results
+  REQUIRE(children.size() == 2);
+
+  REQUIRE(children[0].path().native() == URI(file).to_path());
+  REQUIRE(children[1].path().native() == URI(subdir).to_path());
+
+  REQUIRE(children[0].file_size() == 6);
+
+  // Directories don't get a size
+  REQUIRE(children[1].file_size() == 0);
+
+  // Clean up
+  vfs.remove_dir(URI(path));
+}
