@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2021 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2022 TileDB, Inc.
  * @copyright Copyright (c) 2016 MIT and Intel Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -36,11 +36,12 @@
 
 #include <unordered_map>
 
+#include "tiledb/common/common.h"
 #include "tiledb/common/status.h"
+#include "tiledb/sm/filesystem/uri.h"
 #include "tiledb/sm/filter/filter_pipeline.h"
 #include "tiledb/sm/misc/constants.h"
 #include "tiledb/sm/misc/hilbert.h"
-#include "tiledb/sm/misc/uri.h"
 #include "tiledb/sm/misc/uuid.h"
 
 using namespace tiledb::common;
@@ -72,12 +73,42 @@ class ArraySchema {
   /** Constructor. */
   ArraySchema(ArrayType array_type);
 
+  /** Constructor.
+   * @param uri The URI of the array schema file.
+   * @param version The format version of this array schema.
+   * @param array_type The array type.
+   * @param allows_dups True if the (sparse) array allows coordinate duplicates.
+   * @param domain The array domain.
+   * @param cell_order The cell order.
+   * @param tile_order The tile order.
+   * @param capacity The tile capacity for the case of sparse fragments.
+   * @param attributes The array attributes.
+   * @param cell_var_offsets_filters
+   *    The filter pipeline run on offset tiles for var-length attributes.
+   * @param cell_validity_filters
+   *    The filter pipeline run on validity tiles for nullable attributes.
+   * @param coords_filters The filter pipeline run on coordinate tiles.
+   **/
+  ArraySchema(
+      URI uri,
+      uint32_t version,
+      ArrayType array_type,
+      bool allows_dups,
+      shared_ptr<Domain> domain,
+      Layout cell_order,
+      Layout tile_order,
+      uint64_t capacity,
+      std::vector<shared_ptr<const Attribute>> attributes,
+      FilterPipeline cell_var_offsets_filters,
+      FilterPipeline cell_validity_filters,
+      FilterPipeline coords_filters);
+
   /**
    * Constructor. Clones the input.
    *
    * @param array_schema The array schema to copy.
    */
-  explicit ArraySchema(const ArraySchema* array_schema);
+  explicit ArraySchema(const ArraySchema& array_schema);
 
   /** Destructor. */
   ~ArraySchema();
@@ -114,7 +145,7 @@ class ArraySchema {
   unsigned int attribute_num() const;
 
   /** Returns the attributes. */
-  const std::vector<Attribute*>& attributes() const;
+  const std::vector<shared_ptr<const Attribute>>& attributes() const;
 
   /** Returns the capacity. */
   uint64_t capacity() const;
@@ -163,13 +194,13 @@ class ArraySchema {
   bool dense() const;
 
   /** Returns the i-th dimension. */
-  const Dimension* dimension(unsigned int i) const;
+  shared_ptr<const Dimension> dimension(unsigned int i) const;
 
   /**
    * Returns a constant pointer to the selected dimension (nullptr if it
    * does not exist).
    */
-  const Dimension* dimension(const std::string& name) const;
+  shared_ptr<const Dimension> dimension(const std::string& name) const;
 
   /** Returns the dimension names. */
   std::vector<std::string> dim_names() const;
@@ -239,7 +270,8 @@ class ArraySchema {
    *     they are doing in this case).
    * @return Status
    */
-  Status add_attribute(const Attribute* attr, bool check_special = true);
+  Status add_attribute(
+      shared_ptr<const Attribute> attr, bool check_special = true);
 
   /**
    * Drops an attribute.
@@ -253,12 +285,15 @@ class ArraySchema {
    * It assigns values to the members of the object from the input buffer.
    *
    * @param buff The binary representation of the object to read from.
-   * @return Status
+   * @param uri An optional uri object.
+   * @return A new ArraySchema.
    */
-  Status deserialize(ConstBuffer* buff);
+  static ArraySchema deserialize(ConstBuffer* buff, const URI& uri);
 
   /** Returns the array domain. */
-  const Domain* domain() const;
+  inline const Domain& domain() const {
+    return *const_cast<const Domain*>(domain_.get());
+  }
 
   /**
    * Initializes the ArraySchema object. It also performs a check to see if
@@ -296,7 +331,7 @@ class ArraySchema {
    * Sets the domain. The function returns an error if the array has been
    * previously set to be a key-value store.
    */
-  Status set_domain(Domain* domain);
+  Status set_domain(shared_ptr<Domain> domain);
 
   /** Sets the tile order. */
   Status set_tile_order(Layout tile_order);
@@ -321,19 +356,10 @@ class ArraySchema {
   uint64_t timestamp_start() const;
 
   /** Returns the array schema uri. */
-  URI uri();
-
-  /** Set schema URI, along with parsing out timestamp ranges and name. */
-  void set_uri(const URI& uri);
-
-  /** Get schema URI with return status. */
-  Status get_uri(URI* uri);
+  const URI& uri() const;
 
   /** Returns the schema name. If it is not set, will build it. */
-  std::string name();
-
-  /** Returns the schema name. If it is not set, will returns error status. */
-  Status get_name(std::string* name) const;
+  const std::string& name() const;
 
   /** Set the schema name. */
   void set_name(const std::string& name);
@@ -341,32 +367,48 @@ class ArraySchema {
   /** Generates a new array schema URI. */
   Status generate_uri();
 
+  /** Generates a new array schema URI with specified timestamp range. */
+  Status generate_uri(const std::pair<uint64_t, uint64_t>& timestamp_range);
+
  private:
   /* ********************************* */
   /*         PRIVATE ATTRIBUTES        */
   /* ********************************* */
 
+  /** The URI of the array schema file. */
+  URI uri_;
+
+  /** An array name attached to the schema object. */
+  URI array_uri_;
+
+  /** The format version of this array schema. */
+  uint32_t version_;
+
+  /**
+   * The timestamp the array schema was written.
+   * This is used to determine the array schema file name.
+   * The two timestamps are identical.
+   * It is stored as a pair to keep the usage consistent with metadata
+   */
+  std::pair<uint64_t, uint64_t> timestamp_range_;
+
+  /** The file name of array schema in the format of timestamp_timestamp_uuid */
+  std::string name_;
+
+  /** The array type. */
+  ArrayType array_type_;
+
   /**
    * True if the array allows coordinate duplicates. Applicable only
    * to sparse arrays.
    */
   bool allows_dups_;
 
-  /** An array name attached to the schema object. */
-  URI array_uri_;
+  /** The array domain. */
+  shared_ptr<Domain> domain_;
 
-  /** The array type. */
-  ArrayType array_type_;
-
-  /** It maps each attribute name to the corresponding attribute object. */
-  std::unordered_map<std::string, Attribute*> attribute_map_;
-
-  /** The array attributes. */
-  std::vector<Attribute*> attributes_;
-  /**
-   * The tile capacity for the case of sparse fragments.
-   */
-  uint64_t capacity_;
+  /** It maps each dimension name to the corresponding dimension object. */
+  std::unordered_map<std::string, shared_ptr<const Dimension>> dim_map_;
 
   /**
    * The cell order. It can be one of the following:
@@ -375,6 +417,26 @@ class ArraySchema {
    */
   Layout cell_order_;
 
+  /**
+   * The tile order. It can be one of the following:
+   *    - TILEDB_ROW_MAJOR
+   *    - TILEDB_COL_MAJOR
+   */
+  Layout tile_order_;
+
+  /**
+   * The tile capacity for the case of sparse fragments.
+   */
+  uint64_t capacity_;
+
+  /** It maps each attribute name to the corresponding attribute object.
+   * Lifespan is maintained by the shared_ptr in attributes_. */
+  std::unordered_map<std::string, const Attribute*> attribute_map_;
+
+  /** The array attributes.
+   * Maintains lifespan for elements in both attributes_ and attribute_map_. */
+  std::vector<shared_ptr<const Attribute>> attributes_;
+
   /** The filter pipeline run on offset tiles for var-length attributes. */
   FilterPipeline cell_var_offsets_filters_;
 
@@ -384,39 +446,9 @@ class ArraySchema {
   /** The filter pipeline run on coordinate tiles. */
   FilterPipeline coords_filters_;
 
-  /** It maps each dimension name to the corresponding dimension object. */
-  std::unordered_map<std::string, const Dimension*> dim_map_;
-
-  /** The array domain. */
-  Domain* domain_;
-
-  /**
-   * The tile order. It can be one of the following:
-   *    - TILEDB_ROW_MAJOR
-   *    - TILEDB_COL_MAJOR
-   */
-  Layout tile_order_;
-
-  /** The format version of this array schema. */
-  uint32_t version_;
-
   /** Mutex for thread-safety. */
   mutable std::mutex mtx_;
 
-  /**
-   * The timestamp the array schema was written.
-   * This is used to determine the array schema file name.
-   * The two timestamps are identical.
-   * It is stored as a pair to keep the usage consistent with metadata
-   */
-  std::pair<uint64_t, uint64_t> timestamp_range_;
-
-  /** The URI of the array schema file. */
-  URI uri_;
-
-  /** The file name of array schema in the format of timestamp_timestamp_uuid */
-  std::string name_;
-
   /* ********************************* */
   /*           PRIVATE METHODS         */
   /* ********************************* */
@@ -431,7 +463,14 @@ class ArraySchema {
    * Returns error if double delta compression is used in the zipped
    * coordinate filters and is inherited by a dimension.
    */
-  Status check_double_delta_compressor() const;
+  Status check_double_delta_compressor(
+      const FilterPipeline& coords_filters) const;
+
+  /**
+   * Returns error if RLE or Dictionary encoding is used for string
+   * dimensions but it is not the only filter in the filter list.
+   */
+  Status check_string_compressor(const FilterPipeline& coords_filters) const;
 
   /** Clears all members. Use with caution! */
   void clear();
