@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2021 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2022 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -30,24 +30,28 @@
  * This file implements class Domain.
  */
 
-#include "tiledb/sm/array_schema/domain.h"
+#include "domain.h"
+#include "dimension.h"
+#include "domain_data_ref.h"
+#include "domain_typed_data_view.h"
 #include "tiledb/common/blank.h"
 #include "tiledb/common/heap_memory.h"
 #include "tiledb/common/logger.h"
-#include "tiledb/sm/array_schema/dimension.h"
 #include "tiledb/sm/enums/datatype.h"
 #include "tiledb/sm/enums/layout.h"
+#include "tiledb/sm/misc/tdb_math.h"
 #include "tiledb/sm/misc/utils.h"
+#include "tiledb/type/range/range.h"
 
 #include <cassert>
 #include <iostream>
 #include <limits>
-#include <sstream>
+#include <stdexcept>
 
 using namespace tiledb::common;
+using namespace tiledb::type;
 
-namespace tiledb {
-namespace sm {
+namespace tiledb::sm {
 
 /* ********************************* */
 /*     CONSTRUCTORS & DESTRUCTORS    */
@@ -60,6 +64,27 @@ Domain::Domain() {
   cell_num_per_tile_ = 0;
 }
 
+Domain::Domain(
+    Layout cell_order,
+    const std::vector<shared_ptr<Dimension>> dimensions,
+    Layout tile_order)
+    : cell_order_(cell_order)
+    , dimensions_(dimensions)
+    , dim_num_((unsigned int)dimensions.size())
+    , tile_order_(tile_order) {
+  // Initialize the dimensions mirror
+  dimension_ptrs_.reserve(dimensions_.size());
+  for (const auto& dim : dimensions_) {
+    dimension_ptrs_.emplace_back(dim.get());
+  }
+
+  // Compute number of cells per tile
+  compute_cell_num_per_tile();
+
+  // Compute number of cells per tile
+  set_tile_cell_order_cmp_funcs();
+}
+
 Domain::Domain(const Domain* domain) {
   cell_num_per_tile_ = domain->cell_num_per_tile_;
   cell_order_ = domain->cell_order_;
@@ -68,9 +93,15 @@ Domain::Domain(const Domain* domain) {
   cell_order_cmp_func_2_ = domain->cell_order_cmp_func_2_;
   tile_order_cmp_func_ = domain->tile_order_cmp_func_;
 
-  dimensions_.reserve(domain->dimensions_.size());
-  for (const auto& dim : domain->dimensions_)
-    dimensions_.emplace_back(tdb_new(Dimension, dim.get()));
+  const auto n_dimensions{domain->dimensions_.size()};
+  dimensions_.reserve(n_dimensions);
+  for (const auto& dim : domain->dimensions_) {
+    dimensions_.emplace_back(dim);
+  }
+  dimension_ptrs_.reserve(n_dimensions);
+  for (const auto dim_ptr : domain->dimension_ptrs_) {
+    dimension_ptrs_.emplace_back(dim_ptr);
+  }
 
   tile_order_ = domain->tile_order_;
 }
@@ -78,23 +109,25 @@ Domain::Domain(const Domain* domain) {
 Domain::Domain(Domain&& rhs)
     : cell_num_per_tile_(rhs.cell_num_per_tile_)
     , cell_order_(rhs.cell_order_)
-    , dimensions_(std::move(rhs.dimensions_))
+    , dimensions_(move(rhs.dimensions_))
+    , dimension_ptrs_(move(rhs.dimension_ptrs_))
     , dim_num_(rhs.dim_num_)
     , tile_order_(rhs.tile_order_)
-    , cell_order_cmp_func_(std::move(rhs.cell_order_cmp_func_))
-    , cell_order_cmp_func_2_(std::move(rhs.cell_order_cmp_func_2_))
-    , tile_order_cmp_func_(std::move(rhs.tile_order_cmp_func_)) {
+    , cell_order_cmp_func_(move(rhs.cell_order_cmp_func_))
+    , cell_order_cmp_func_2_(move(rhs.cell_order_cmp_func_2_))
+    , tile_order_cmp_func_(move(rhs.tile_order_cmp_func_)) {
 }
 
 Domain& Domain::operator=(Domain&& rhs) {
   cell_num_per_tile_ = rhs.cell_num_per_tile_;
   cell_order_ = rhs.cell_order_;
   dim_num_ = rhs.dim_num_;
-  cell_order_cmp_func_ = std::move(rhs.cell_order_cmp_func_);
-  tile_order_cmp_func_ = std::move(rhs.tile_order_cmp_func_);
-  dimensions_ = std::move(rhs.dimensions_);
+  cell_order_cmp_func_ = move(rhs.cell_order_cmp_func_);
+  tile_order_cmp_func_ = move(rhs.tile_order_cmp_func_);
+  dimensions_ = move(rhs.dimensions_);
+  dimension_ptrs_ = move(rhs.dimension_ptrs_);
   tile_order_ = rhs.tile_order_;
-  cell_order_cmp_func_2_ = std::move(rhs.cell_order_cmp_func_2_);
+  cell_order_cmp_func_2_ = move(rhs.cell_order_cmp_func_2_);
 
   return *this;
 }
@@ -111,15 +144,16 @@ Layout Domain::tile_order() const {
   return tile_order_;
 }
 
-Status Domain::add_dimension(const Dimension* dim) {
-  dimensions_.emplace_back(tdb_new(Dimension, dim));
+Status Domain::add_dimension(shared_ptr<Dimension> dim) {
+  dimensions_.emplace_back(dim);
+  dimension_ptrs_.emplace_back(dim.get());
   ++dim_num_;
 
   return Status::Ok();
 }
 
 bool Domain::all_dims_fixed() const {
-  for (const auto& dim : dimensions_) {
+  for (const auto dim : dimension_ptrs_) {
     if (dim->var_size())
       return false;
   }
@@ -128,7 +162,7 @@ bool Domain::all_dims_fixed() const {
 }
 
 bool Domain::all_dims_int() const {
-  for (const auto& dim : dimensions_) {
+  for (const auto dim : dimension_ptrs_) {
     if (!datatype_is_integer(dim->type()))
       return false;
   }
@@ -137,7 +171,7 @@ bool Domain::all_dims_int() const {
 }
 
 bool Domain::all_dims_string() const {
-  for (const auto& dim : dimensions_) {
+  for (const auto dim : dimension_ptrs_) {
     if (!datatype_is_string(dim->type()))
       return false;
   }
@@ -146,7 +180,7 @@ bool Domain::all_dims_string() const {
 }
 
 bool Domain::all_dims_real() const {
-  for (const auto& dim : dimensions_) {
+  for (const auto dim : dimension_ptrs_) {
     if (!datatype_is_real(dim->type()))
       return false;
   }
@@ -155,12 +189,12 @@ bool Domain::all_dims_real() const {
 }
 
 bool Domain::all_dims_same_type() const {
-  if (dim_num_ == 0)
+  if (dim_num_ <= 1)
     return true;
 
-  auto type = dimensions_[0]->type();
+  auto type = dimension_ptrs_[0]->type();
   for (unsigned d = 1; d < dim_num_; ++d) {
-    if (dimensions_[d]->type() != type)
+    if (dimension_ptrs_[d]->type() != type)
       return false;
   }
 
@@ -172,30 +206,16 @@ uint64_t Domain::cell_num_per_tile() const {
 }
 
 template <>
-int Domain::cell_order_cmp<char>(
-    const Dimension* dim, const QueryBuffer* buff, uint64_t a, uint64_t b) {
+int Domain::cell_order_cmp_impl<char>(
+    const Dimension* dim, const UntypedDatumView a, const UntypedDatumView b) {
   // Must be var-sized
   assert(dim->var_size());
   (void)dim;
 
-  auto offs = (const uint64_t*)buff->buffer_;
-  auto offs_size = *(buff->buffer_size_);
-  auto var_size = *(buff->buffer_var_size_);
-  auto off_a = offs[a];
-  auto off_b = offs[b];
-  auto var_a = &((const char*)buff->buffer_var_)[off_a];
-  auto var_b = &((const char*)buff->buffer_var_)[off_b];
-
-  // If coordinates are last, the next offset is the size of the variable buffer
-  // Otherwisse, it is just the next offset in the fixed buffer
-  auto next_off_a = ((a + 1) * constants::cell_var_offset_size == offs_size) ?
-                        var_size :
-                        offs[a + 1];
-  auto next_off_b = ((b + 1) * constants::cell_var_offset_size == offs_size) ?
-                        var_size :
-                        offs[b + 1];
-  auto size_a = next_off_a - off_a;
-  auto size_b = next_off_b - off_b;
+  auto var_a{a.value_as<const char[]>()};
+  auto var_b{b.value_as<const char[]>()};
+  auto size_a{a.size()};
+  auto size_b{b.size()};
   auto size = std::min(size_a, size_b);
 
   // Check common prefix of size `size`
@@ -215,14 +235,13 @@ int Domain::cell_order_cmp<char>(
 }
 
 template <class T>
-int Domain::cell_order_cmp(
-    const Dimension* dim, const QueryBuffer* buff, uint64_t a, uint64_t b) {
-  // Must be fixed-sized
+int Domain::cell_order_cmp_impl(
+    const Dimension* dim, const UntypedDatumView a, const UntypedDatumView b) {
   assert(!dim->var_size());
   (void)dim;
 
-  auto ca = ((const T*)buff->buffer_)[a];
-  auto cb = ((const T*)buff->buffer_)[b];
+  auto ca = a.template value_as<T>();
+  auto cb = b.template value_as<T>();
   if (ca < cb)
     return -1;
   if (ca > cb)
@@ -231,11 +250,11 @@ int Domain::cell_order_cmp(
 }
 
 int Domain::cell_order_cmp(
-    unsigned dim_idx, const ResultCoords& a, const ResultCoords& b) const {
+    unsigned dim_idx, UntypedDatumView a, UntypedDatumView b) const {
   // Handle variable-sized dimensions
-  if (dimensions_[dim_idx]->var_size()) {
-    auto s_a = a.coord_string(dim_idx);
-    auto s_b = b.coord_string(dim_idx);
+  if (dimension_ptrs_[dim_idx]->var_size()) {
+    std::string_view s_a{static_cast<const char*>(a.content()), a.size()};
+    std::string_view s_b{static_cast<const char*>(b.content()), b.size()};
 
     if (s_a == s_b)
       return 0;
@@ -244,10 +263,11 @@ int Domain::cell_order_cmp(
     return 1;
   }
 
-  assert(cell_order_cmp_func_2_[dim_idx] != nullptr);
-  auto coord_a = a.coord(dim_idx);
-  auto coord_b = b.coord(dim_idx);
-  return cell_order_cmp_func_2_[dim_idx](coord_a, coord_b);
+  if (cell_order_cmp_func_2_[dim_idx] == nullptr) {
+    assert(cell_order_cmp_func_2_[dim_idx] != nullptr);
+    throw std::logic_error("comparison function not initialized");
+  }
+  return cell_order_cmp_func_2_[dim_idx](a.content(), b.content());
 }
 
 template <class T>
@@ -262,13 +282,12 @@ int Domain::cell_order_cmp_2(const void* coord_a, const void* coord_b) {
 }
 
 int Domain::cell_order_cmp(
-    const std::vector<const QueryBuffer*>& coord_buffs,
-    uint64_t a,
-    uint64_t b) const {
+    const type::DomainDataRef& left, const type::DomainDataRef& right) const {
   if (cell_order_ == Layout::ROW_MAJOR) {
     for (unsigned d = 0; d < dim_num_; ++d) {
-      auto dim = dimension(d);
-      auto res = cell_order_cmp_func_[d](dim, coord_buffs[d], a, b);
+      auto dim = dimension_ptr(d);
+      auto res = cell_order_cmp_func_[d](
+          dim, left.dimension_datum_view(d), right.dimension_datum_view(d));
 
       if (res == 1 || res == -1)
         return res;
@@ -276,8 +295,9 @@ int Domain::cell_order_cmp(
     }
   } else {  // COL_MAJOR
     for (unsigned d = dim_num_ - 1;; --d) {
-      auto dim = dimension(d);
-      auto res = cell_order_cmp_func_[d](dim, coord_buffs[d], a, b);
+      auto dim = dimension_ptr(d);
+      auto res = cell_order_cmp_func_[d](
+          dim, left.dimension_datum_view(d), right.dimension_datum_view(d));
 
       if (res == 1 || res == -1)
         return res;
@@ -293,59 +313,68 @@ int Domain::cell_order_cmp(
 
 void Domain::crop_ndrange(NDRange* ndrange) const {
   for (unsigned d = 0; d < dim_num_; ++d)
-    dimensions_[d]->crop_range(&(*ndrange)[d]);
+    dimension_ptrs_[d]->crop_range(&(*ndrange)[d]);
 }
 
-Status Domain::deserialize(ConstBuffer* buff, uint32_t version) {
+tuple<Status, optional<shared_ptr<Domain>>> Domain::deserialize(
+    ConstBuffer* buff, uint32_t version, Layout cell_order, Layout tile_order) {
+  Status st;
   // Load type
   Datatype type = Datatype::INT32;
   if (version < 5) {
     uint8_t type_c;
-    RETURN_NOT_OK(buff->read(&type_c, sizeof(uint8_t)));
+    st = buff->read(&type_c, sizeof(uint8_t));
+    if (!st.ok()) {
+      return {st, nullopt};
+    }
     type = static_cast<Datatype>(type_c);
   }
 
+  std::vector<shared_ptr<Dimension>> dimensions;
+  uint32_t dim_num;
   // Load dimensions
-  RETURN_NOT_OK(buff->read(&dim_num_, sizeof(uint32_t)));
-  for (uint32_t i = 0; i < dim_num_; ++i) {
-    auto dim = tdb_new(tiledb::common::blank<Dimension>);
-    dim->deserialize(buff, version, type);
-    dimensions_.emplace_back(dim);
+  st = buff->read(&dim_num, sizeof(uint32_t));
+  if (!st.ok()) {
+    return {st, nullopt};
+  }
+  for (uint32_t i = 0; i < dim_num; ++i) {
+    auto&& [st_dim, dim]{Dimension::deserialize(buff, version, type)};
+    if (!st_dim.ok()) {
+      return {Status_DomainError("Cannot deserialize dimension."), nullopt};
+    }
+    dimensions.emplace_back(std::move(dim.value()));
   }
 
-  set_tile_cell_order_cmp_funcs();
-
-  return Status::Ok();
-}
-
-unsigned int Domain::dim_num() const {
-  return dim_num_;
+  return {Status::Ok(),
+          tiledb::common::make_shared<Domain>(
+              HERE(), cell_order, dimensions, tile_order)};
 }
 
 const Range& Domain::domain(unsigned i) const {
   assert(i < dim_num_);
-  return dimensions_[i]->domain();
+  return dimension_ptrs_[i]->domain();
 }
 
 NDRange Domain::domain() const {
   NDRange ret(dim_num_);
   for (unsigned d = 0; d < dim_num_; ++d)
-    ret[d] = dimensions_[d]->domain();
+    ret[d] = dimension_ptrs_[d]->domain();
 
   return ret;
 }
 
-const Dimension* Domain::dimension(unsigned int i) const {
-  if (i > dim_num_)
+shared_ptr<const Dimension> Domain::dimension(unsigned int i) const {
+  if (i > dim_num_) {
     return nullptr;
-  return dimensions_[i].get();
+  }
+  return dimensions_[i];
 }
 
-const Dimension* Domain::dimension(const std::string& name) const {
+shared_ptr<const Dimension> Domain::dimension(const std::string& name) const {
   for (unsigned int i = 0; i < dim_num_; i++) {
     const auto& dim = dimensions_[i];
     if (dim->name() == name) {
-      return dim.get();
+      return dim;
     }
   }
   return nullptr;
@@ -355,7 +384,7 @@ void Domain::dump(FILE* out) const {
   if (out == nullptr)
     out = stdout;
 
-  for (const auto& dim : dimensions_) {
+  for (const auto dim : dimension_ptrs_) {
     fprintf(out, "\n");
     dim->dump(out);
   }
@@ -372,7 +401,7 @@ void Domain::expand_ndrange(const NDRange& r1, NDRange* r2) const {
 
   // Expand r2 along all dimensions
   for (unsigned d = 0; d < dim_num_; ++d) {
-    const auto& dim = dimensions_[d];
+    const auto dim = dimension_ptrs_[d];
     if (!dim->var_size())
       dim->expand_range(r1[d], &(*r2)[d]);
     else
@@ -382,10 +411,11 @@ void Domain::expand_ndrange(const NDRange& r1, NDRange* r2) const {
 
 void Domain::expand_to_tiles(NDRange* ndrange) const {
   for (unsigned d = 0; d < dim_num_; ++d) {
-    const auto& dim = dimensions_[d];
+    const auto dim = dimension_ptrs_[d];
     // Applicable only to fixed-sized dimensions
-    if (!dim->var_size())
-      dimensions_[d]->expand_to_tile(&(*ndrange)[d]);
+    if (!dim->var_size()) {
+      dim->expand_to_tile(&(*ndrange)[d]);
+    }
   }
 }
 
@@ -501,7 +531,7 @@ void Domain::get_tile_subarray(
 Status Domain::has_dimension(const std::string& name, bool* has_dim) const {
   *has_dim = false;
 
-  for (const auto& dim : dimensions_) {
+  for (const auto dim : dimension_ptrs_) {
     if (name == dim->name()) {
       *has_dim = true;
       break;
@@ -514,13 +544,13 @@ Status Domain::has_dimension(const std::string& name, bool* has_dim) const {
 Status Domain::get_dimension_index(
     const std::string& name, unsigned* dim_idx) const {
   for (unsigned d = 0; d < dim_num_; ++d) {
-    if (dimensions_[d]->name() == name) {
+    if (dimension_ptrs_[d]->name() == name) {
       *dim_idx = d;
       return Status::Ok();
     }
   }
 
-  return Status::DomainError(
+  return Status_DomainError(
       "Cannot get dimension index; Invalid dimension name");
 }
 
@@ -597,7 +627,7 @@ uint64_t Domain::stride(Layout subarray_layout) const {
 
 const ByteVecValue& Domain::tile_extent(unsigned i) const {
   assert(i < dim_num_);
-  return dimensions_[i]->tile_extent();
+  return dimension_ptrs_[i]->tile_extent();
 }
 
 std::vector<ByteVecValue> Domain::tile_extents() const {
@@ -611,7 +641,7 @@ std::vector<ByteVecValue> Domain::tile_extents() const {
 uint64_t Domain::tile_num(const NDRange& ndrange) const {
   uint64_t ret = 1;
   for (unsigned d = 0; d < dim_num_; ++d)
-    ret *= dimensions_[d]->tile_num(ndrange[d]);
+    ret *= dimension_ptrs_[d]->tile_num(ndrange[d]);
 
   return ret;
 }
@@ -620,7 +650,7 @@ uint64_t Domain::cell_num(const NDRange& ndrange) const {
   assert(!ndrange.empty());
   uint64_t cell_num = 1, range;
   for (unsigned d = 0; d < dim_num_; ++d) {
-    range = dimensions_[d]->domain_range(ndrange[d]);
+    range = dimension_ptrs_[d]->domain_range(ndrange[d]);
     if (range == std::numeric_limits<uint64_t>::max())  // Overflow
       return range;
 
@@ -637,7 +667,7 @@ bool Domain::covered(const NDRange& r1, const NDRange& r2) const {
   assert(r2.size() == dim_num_);
 
   for (unsigned d = 0; d < dim_num_; ++d) {
-    if (!dimensions_[d]->covered(r1[d], r2[d]))
+    if (!dimension_ptrs_[d]->covered(r1[d], r2[d]))
       return false;
   }
 
@@ -649,23 +679,29 @@ bool Domain::overlap(const NDRange& r1, const NDRange& r2) const {
   assert(r2.size() == dim_num_);
 
   for (unsigned d = 0; d < dim_num_; ++d) {
-    if (!dimensions_[d]->overlap(r1[d], r2[d]))
+    if (!dimension_ptrs_[d]->overlap(r1[d], r2[d]))
       return false;
   }
 
   return true;
 }
 
-double Domain::overlap_ratio(const NDRange& r1, const NDRange& r2) const {
+double Domain::overlap_ratio(
+    const NDRange& r1,
+    const std::vector<bool>& r1_default,
+    const NDRange& r2) const {
   double ratio = 1.0;
   assert(dim_num_ == r1.size());
   assert(dim_num_ == r2.size());
 
   for (unsigned d = 0; d < dim_num_; ++d) {
-    if (!dimensions_[d]->overlap(r1[d], r2[d]))
+    if (r1_default[d])
+      continue;
+
+    if (!dimension_ptrs_[d]->overlap(r1[d], r2[d]))
       return 0.0;
 
-    ratio *= dimensions_[d]->overlap_ratio(r1[d], r2[d]);
+    ratio *= dimension_ptrs_[d]->overlap_ratio(r1[d], r2[d]);
 
     // If ratio goes to 0, then the subarray overlap is much smaller than the
     // volume of the MBR. Since we have already guaranteed that there is an
@@ -679,7 +715,7 @@ double Domain::overlap_ratio(const NDRange& r1, const NDRange& r2) const {
 }
 
 template <class T>
-int Domain::tile_order_cmp(
+int Domain::tile_order_cmp_impl(
     const Dimension* dim, const void* coord_a, const void* coord_b) {
   if (!dim->tile_extent())
     return 0;
@@ -698,21 +734,19 @@ int Domain::tile_order_cmp(
 }
 
 int Domain::tile_order_cmp(
-    const std::vector<const QueryBuffer*>& coord_buffs,
-    uint64_t a,
-    uint64_t b) const {
+    const type::DomainDataRef& left, const type::DomainDataRef& right) const {
   if (tile_order_ == Layout::ROW_MAJOR) {
     for (unsigned d = 0; d < dim_num_; ++d) {
-      auto dim = dimension(d);
+      auto dim = dimension_ptr(d);
 
       // Inapplicable to var-sized dimensions or absent tile extents
       if (dim->var_size() || !dim->tile_extent())
         continue;
 
-      auto coord_size = dim->coord_size();
-      auto ca = &(((unsigned char*)coord_buffs[d]->buffer_)[a * coord_size]);
-      auto cb = &(((unsigned char*)coord_buffs[d]->buffer_)[b * coord_size]);
-      auto res = tile_order_cmp_func_[d](dim, ca, cb);
+      auto res = tile_order_cmp_func_[d](
+          dim,
+          left.dimension_datum_view(d).content(),
+          right.dimension_datum_view(d).content());
 
       if (res == 1 || res == -1)
         return res;
@@ -720,14 +754,14 @@ int Domain::tile_order_cmp(
     }
   } else {  // COL_MAJOR
     for (unsigned d = dim_num_ - 1;; --d) {
-      auto dim = dimension(d);
+      auto dim = dimension_ptr(d);
 
       // Inapplicable to var-sized dimensions or absent tile extents
       if (!dim->var_size() && dim->tile_extent()) {
-        auto coord_size = dim->coord_size();
-        auto ca = &(((unsigned char*)coord_buffs[d]->buffer_)[a * coord_size]);
-        auto cb = &(((unsigned char*)coord_buffs[d]->buffer_)[b * coord_size]);
-        auto res = tile_order_cmp_func_[d](dim, ca, cb);
+        auto res = tile_order_cmp_func_[d](
+            dim,
+            left.dimension_datum_view(d).content(),
+            right.dimension_datum_view(d).content());
 
         if (res == 1 || res == -1)
           return res;
@@ -745,7 +779,7 @@ int Domain::tile_order_cmp(
 int Domain::tile_order_cmp(
     unsigned dim_idx, const void* coord_a, const void* coord_b) const {
   auto dim = dimension(dim_idx);
-  return tile_order_cmp_func_[dim_idx](dim, coord_a, coord_b);
+  return tile_order_cmp_func_[dim_idx](dim.get(), coord_a, coord_b);
 }
 
 /* ****************************** */
@@ -758,7 +792,7 @@ void Domain::compute_cell_num_per_tile() {
     return;
 
   // Invoke the proper templated function
-  auto type = dimensions_[0]->type();
+  auto type{dimension_ptrs_[0]->type()};
   switch (type) {
     case Datatype::INT32:
       compute_cell_num_per_tile<int>();
@@ -836,46 +870,46 @@ void Domain::set_tile_cell_order_cmp_funcs() {
   cell_order_cmp_func_.resize(dim_num_);
   cell_order_cmp_func_2_.resize(dim_num_);
   for (unsigned d = 0; d < dim_num_; ++d) {
-    auto type = dimensions_[d]->type();
+    auto type{dimension_ptrs_[d]->type()};
     switch (type) {
       case Datatype::INT32:
-        tile_order_cmp_func_[d] = tile_order_cmp<int32_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<int32_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<int32_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<int32_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<int32_t>;
         break;
       case Datatype::INT64:
-        tile_order_cmp_func_[d] = tile_order_cmp<int64_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<int64_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<int64_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<int64_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<int64_t>;
         break;
       case Datatype::INT8:
-        tile_order_cmp_func_[d] = tile_order_cmp<int8_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<int8_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<int8_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<int8_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<int8_t>;
         break;
       case Datatype::UINT8:
-        tile_order_cmp_func_[d] = tile_order_cmp<uint8_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<uint8_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<uint8_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<uint8_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<uint8_t>;
         break;
       case Datatype::INT16:
-        tile_order_cmp_func_[d] = tile_order_cmp<int16_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<int16_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<int16_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<int16_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<int16_t>;
         break;
       case Datatype::UINT16:
-        tile_order_cmp_func_[d] = tile_order_cmp<uint16_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<uint16_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<uint16_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<uint16_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<uint16_t>;
         break;
       case Datatype::UINT32:
-        tile_order_cmp_func_[d] = tile_order_cmp<uint32_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<uint32_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<uint32_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<uint32_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<uint32_t>;
         break;
       case Datatype::UINT64:
-        tile_order_cmp_func_[d] = tile_order_cmp<uint64_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<uint64_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<uint64_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<uint64_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<uint64_t>;
         break;
       case Datatype::DATETIME_YEAR:
@@ -900,25 +934,26 @@ void Domain::set_tile_cell_order_cmp_funcs() {
       case Datatype::TIME_PS:
       case Datatype::TIME_FS:
       case Datatype::TIME_AS:
-        tile_order_cmp_func_[d] = tile_order_cmp<int64_t>;
-        cell_order_cmp_func_[d] = cell_order_cmp<int64_t>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<int64_t>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<int64_t>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<int64_t>;
         break;
       case Datatype::FLOAT32:
-        tile_order_cmp_func_[d] = tile_order_cmp<float>;
-        cell_order_cmp_func_[d] = cell_order_cmp<float>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<float>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<float>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<float>;
         break;
       case Datatype::FLOAT64:
-        tile_order_cmp_func_[d] = tile_order_cmp<double>;
-        cell_order_cmp_func_[d] = cell_order_cmp<double>;
+        tile_order_cmp_func_[d] = tile_order_cmp_impl<double>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<double>;
         cell_order_cmp_func_2_[d] = cell_order_cmp_2<double>;
         break;
       case Datatype::STRING_ASCII:
         tile_order_cmp_func_[d] = nullptr;
-        cell_order_cmp_func_[d] = cell_order_cmp<char>;
+        cell_order_cmp_func_[d] = cell_order_cmp_impl<char>;
         cell_order_cmp_func_2_[d] = nullptr;
         break;
+      case Datatype::BLOB:
       case Datatype::CHAR:
       case Datatype::STRING_UTF8:
       case Datatype::STRING_UTF16:
@@ -1217,5 +1252,4 @@ template uint64_t Domain::stride<uint64_t>(Layout subarray_layout) const;
 template uint64_t Domain::stride<float>(Layout subarray_layout) const;
 template uint64_t Domain::stride<double>(Layout subarray_layout) const;
 
-}  // namespace sm
-}  // namespace tiledb
+}  // namespace tiledb::sm
