@@ -44,6 +44,7 @@
 #include "tiledb/sm/filter/compression_filter.h"
 #include "tiledb/sm/filter/filter_pipeline.h"
 #include "tiledb/sm/fragment/fragment_info.h"
+#include "tiledb/sm/group/group.h"
 #include "tiledb/sm/query/query.h"
 #include "tiledb/sm/query/query_condition.h"
 #include "tiledb/sm/storage_manager/context.h"
@@ -54,6 +55,10 @@ struct tiledb_array_t {
   tiledb::sm::Array* array_ = nullptr;
 };
 
+struct tiledb_subarray_t {
+  tiledb::sm::Subarray* subarray_ = nullptr;
+};
+
 struct tiledb_buffer_t {
   tiledb::sm::Datatype datatype_ = tiledb::sm::Datatype::UINT8;
   tiledb::sm::Buffer* buffer_ = nullptr;
@@ -84,7 +89,7 @@ struct tiledb_attribute_t {
 };
 
 struct tiledb_array_schema_t {
-  tiledb::sm::ArraySchema* array_schema_ = nullptr;
+  shared_ptr<tiledb::sm::ArraySchema> array_schema_;
 };
 
 struct tiledb_array_schema_evolution_t {
@@ -127,4 +132,8 @@ struct tiledb_fragment_info_t {
   tiledb::sm::FragmentInfo* fragment_info_ = nullptr;
 };
 
+struct tiledb_group_t {
+  tdb_unique_ptr<tiledb::sm::Group> group_ = nullptr;
+};
+
 #endif
