@@ -82,11 +82,25 @@ inline Status layout_enum(const std::string& layout_str, Layout* layout) {
   else if (layout_str == constants::hilbert_str)
     *layout = Layout::HILBERT;
   else {
-    return Status::Error("Invalid Layout " + layout_str);
+    return Status_Error("Invalid Layout " + layout_str);
   }
   return Status::Ok();
 }
 
+inline Status tile_order_is_valid(uint8_t layout_enum) {
+  if (layout_enum != 0 && layout_enum != 1)
+    return Status_Error("[Tile order] Invalid Layout enum.");
+
+  return Status::Ok();
+}
+
+inline Status cell_order_is_valid(uint8_t layout_enum) {
+  if (layout_enum > 4)
+    return Status_Error("[Cell order] Invalid Layout enum.");
+
+  return Status::Ok();
+}
+
 }  // namespace sm
 }  // namespace tiledb
 
