@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2021 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2022 TileDB, Inc.
  * @copyright Copyright (c) 2016 MIT and Intel Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -34,16 +34,23 @@
 #ifndef TILEDB_FRAGMENT_METADATA_H
 #define TILEDB_FRAGMENT_METADATA_H
 
+#include <deque>
 #include <mutex>
 #include <unordered_map>
 #include <vector>
 
+#include "tiledb/common/common.h"
 #include "tiledb/common/status.h"
+#include "tiledb/sm/filesystem/uri.h"
 #include "tiledb/sm/misc/types.h"
-#include "tiledb/sm/misc/uri.h"
 #include "tiledb/sm/rtree/rtree.h"
 
 using namespace tiledb::common;
+using namespace tiledb::type;
+
+namespace tiledb::type {
+class Range;
+}
 
 namespace tiledb {
 namespace sm {
@@ -51,7 +58,7 @@ namespace sm {
 class ArraySchema;
 class Buffer;
 class EncryptionKey;
-class OpenArrayMemoryTracker;
+class MemoryTracker;
 class StorageManager;
 
 /** Stores the metadata structures of a fragment. */
@@ -61,15 +68,15 @@ class FragmentMetadata {
   /*     CONSTRUCTORS & DESTRUCTORS    */
   /* ********************************* */
 
-  /**
-   * Default contructor
-   */
-  FragmentMetadata() = default;
+  /** Constructor. */
+  FragmentMetadata();
 
   /**
    * Constructor.
    *
    * @param storage_manager A storage manager instance.
+   * @param memory_tracker The memory tracker of the array this fragment
+   *     metadata corresponds to.
    * @param array_schema The schema of the array the fragment belongs to.
    * @param fragment_uri The fragment URI.
    * @param timestamp_range The timestamp range of the fragment.
@@ -79,7 +86,8 @@ class FragmentMetadata {
    */
   FragmentMetadata(
       StorageManager* storage_manager,
-      const ArraySchema* array_schema,
+      MemoryTracker* memory_tracker,
+      const shared_ptr<const ArraySchema>& array_schema,
       const URI& fragment_uri,
       const std::pair<uint64_t, uint64_t>& timestamp_range,
       bool dense = true);
@@ -102,6 +110,12 @@ class FragmentMetadata {
   /** Returns the number of cells in the tile at the input position. */
   uint64_t cell_num(uint64_t tile_pos) const;
 
+  /**
+   * Returns the dimensions types from the array schema asssociated
+   * with this fragment metadata.
+   */
+  std::vector<Datatype> dim_types() const;
+
   /**
    * Computes an upper bound on the buffer sizes needed when reading a subarray
    * from the fragment, for a given set of attributes. Note that these upper
@@ -209,16 +223,12 @@ class FragmentMetadata {
   bool has_consolidated_footer() const;
 
   /**
-   * Returns true if the input range overlaps the non-empty
-   * domain of the fragment.
+   * Retrieves the overlap of all MBRs with the input ND range.
    */
-  bool overlaps_non_empty_domain(const NDRange& range) const;
-
-  /**
-   * Retrieves the overlap of all MBRs with the input ND range. The encryption
-   * key is needed because certain metadata may have to be loaded on-the-fly.
-   */
-  Status get_tile_overlap(const NDRange& range, TileOverlap* tile_overlap);
+  Status get_tile_overlap(
+      const NDRange& range,
+      std::vector<bool>& is_default,
+      TileOverlap* tile_overlap);
 
   /**
    * Compute tile bitmap for the curent fragment/range/dimension.
@@ -246,12 +256,32 @@ class FragmentMetadata {
       const EncryptionKey& encryption_key,
       Buffer* f_buff,
       uint64_t offset,
-      std::unordered_map<std::string, tiledb_shared_ptr<ArraySchema>>
-          array_schemas);
+      std::unordered_map<std::string, shared_ptr<ArraySchema>> array_schemas);
 
   /** Stores all the metadata to storage. */
   Status store(const EncryptionKey& encryption_key);
 
+  /**
+   * Stores all the metadata to storage.
+   *
+   * Applicable to format versions 7 to 10.
+   */
+  Status store_v7_v10(const EncryptionKey& encryption_key);
+
+  /**
+   * Stores all the metadata to storage.
+   *
+   * Applicable to format versions 11.
+   */
+  Status store_v11(const EncryptionKey& encryption_key);
+
+  /**
+   * Stores all the metadata to storage.
+   *
+   * Applicable to format versions 12 or higher.
+   */
+  Status store_v12_or_higher(const EncryptionKey& encryption_key);
+
   /** Returns the non-empty domain in which the fragment is constrained. */
   const NDRange& non_empty_domain();
 
@@ -339,13 +369,115 @@ class FragmentMetadata {
   void set_tile_validity_offset(
       const std::string& name, uint64_t tid, uint64_t step);
 
+  /**
+   * Sets a tile min for the fixed input attribute.
+   *
+   * @param name The attribute for which the min is set.
+   * @param tid The index of the tile for which the min is set.
+   * @param min The minimum.
+   * @param size The size.
+   * @return void
+   */
+  void set_tile_min(
+      const std::string& name, uint64_t tid, const void* min, uint64_t size);
+
+  /**
+   * Sets a tile min size for the var input attribute.
+   *
+   * @param name The attribute for which the min size is set.
+   * @param tid The index of the tile for which the min is set.
+   * @param size The size.
+   * @return void
+   */
+  void set_tile_min_var_size(
+      const std::string& name, uint64_t tid, uint64_t size);
+
+  /**
+   * Sets a tile min for the var input attribute.
+   *
+   * @param name The attribute for which the min is set.
+   * @param tid The index of the tile for which the min is set.
+   * @param min The minimum.
+   * @return void
+   */
+  void set_tile_min_var(const std::string& name, uint64_t tid, const void* min);
+
+  /**
+   * Sets a tile max for the input attribute.
+   *
+   * @param name The attribute for which the max is set.
+   * @param tid The index of the tile for which the max is set.
+   * @param max The maximum.
+   * @param size The size.
+   * @return void
+   */
+  void set_tile_max(
+      const std::string& name, uint64_t tid, const void* max, uint64_t size);
+
+  /**
+   * Sets a tile max for the var input attribute.
+   *
+   * @param name The attribute for which the min size is set.
+   * @param tid The index of the tile for which the min is set.
+   * @param size The size.
+   * @return void
+   */
+  void set_tile_max_var_size(
+      const std::string& name, uint64_t tid, uint64_t size);
+
+  /**
+   * Sets a tile max for the var input attribute.
+   *
+   * @param name The attribute for which the min is set.
+   * @param tid The index of the tile for which the min is set.
+   * @param max The maximum.
+   * @return void
+   */
+  void set_tile_max_var(const std::string& name, uint64_t tid, const void* max);
+
+  /**
+   * Converts min/max sizes to offsets.
+   *
+   * @param name The attribute for which the offsets are converted
+   * @return void
+   */
+  void convert_tile_min_max_var_sizes_to_offsets(const std::string& name);
+
+  /**
+   * Sets a tile sum for the input attribute.
+   *
+   * @param name The attribute for which the sum is set.
+   * @param tid The index of the tile for which the sum is set.
+   * @param sum The sum.
+   * @return void
+   */
+  void set_tile_sum(const std::string& name, uint64_t tid, const ByteVec* sum);
+
+  /**
+   * Sets a tile null count for the input attribute.
+   *
+   * @param name The attribute for which the null count is set.
+   * @param tid The index of the tile for which the null count is set.
+   * @param sum The null count.
+   * @return void
+   */
+  void set_tile_null_count(
+      const std::string& name, uint64_t tid, uint64_t null_count);
+
+  /**
+   * Compute fragment min, max, sum, null count for all dimensions/attributes.
+   *
+   * @return Status.
+   */
+  Status compute_fragment_min_max_sum_null_count();
+
   /**
    * Sets array schema pointer.
    *
    * @param array_schema The schema pointer.
    * @return void
    */
-  void set_array_schema(ArraySchema* array_schema);
+  void set_array_schema(const shared_ptr<const ArraySchema>& array_schema);
 
   /** Returns the tile index base value. */
   uint64_t tile_index_base() const;
@@ -354,13 +486,13 @@ class FragmentMetadata {
   uint64_t tile_num() const;
 
   /** Returns the URI of the input attribute/dimension. */
-  URI uri(const std::string& name) const;
+  tuple<Status, optional<URI>> uri(const std::string& name) const;
 
   /** Returns the URI of the input variable-sized attribute/dimension. */
-  URI var_uri(const std::string& name) const;
+  tuple<Status, optional<URI>> var_uri(const std::string& name) const;
 
   /** Returns the validity URI of the input nullable attribute. */
-  URI validity_uri(const std::string& name) const;
+  tuple<Status, optional<URI>> validity_uri(const std::string& name) const;
 
   /** Return the array schema name. */
   const std::string& array_schema_name();
@@ -424,11 +556,10 @@ class FragmentMetadata {
    *
    * @param name The input attribute/dimension.
    * @param tile_idx The index of the tile in the metadata.
-   * @param tile_size The tile size to be retrieved.
-   * @return Status
+   * @return Status, size
    */
-  Status persisted_tile_size(
-      const std::string& name, uint64_t tile_idx, uint64_t* tile_size);
+  tuple<Status, optional<uint64_t>> persisted_tile_size(
+      const std::string& name, uint64_t tile_idx);
 
   /**
    * Retrieves the size of the tile when it is persisted (e.g. the size of the
@@ -437,11 +568,10 @@ class FragmentMetadata {
    *
    * @param name The input attribute/dimension.
    * @param tile_idx The index of the tile in the metadata.
-   * @param tile_size The tile size to be retrieved.
-   * @return Status
+   * @return Status, size
    */
-  Status persisted_tile_var_size(
-      const std::string& name, uint64_t tile_idx, uint64_t* tile_size);
+  tuple<Status, optional<uint64_t>> persisted_tile_var_size(
+      const std::string& name, uint64_t tile_idx);
 
   /**
    * Retrieves the size of the validity tile when it is persisted (e.g. the size
@@ -449,11 +579,10 @@ class FragmentMetadata {
    *
    * @param name The input attribute.
    * @param tile_idx The index of the tile in the metadata.
-   * @param tile_size The tile size to be retrieved.
-   * @return Status
+   * @return Status, size
    */
-  Status persisted_tile_validity_size(
-      const std::string& name, uint64_t tile_idx, uint64_t* tile_size);
+  tuple<Status, optional<uint64_t>> persisted_tile_validity_size(
+      const std::string& name, uint64_t tile_idx);
 
   /**
    * Returns the (uncompressed) tile size for a given attribute or dimension
@@ -472,11 +601,88 @@ class FragmentMetadata {
    *
    * @param name The input attribute/dimension.
    * @param tile_idx The index of the tile in the metadata.
-   * @param tile_size The tile size to be retrieved.
-   * @return Status
+   * @return Status, size.
    */
-  Status tile_var_size(
-      const std::string& name, uint64_t tile_idx, uint64_t* tile_size);
+  tuple<Status, optional<uint64_t>> tile_var_size(
+      const std::string& name, uint64_t tile_idx);
+
+  /**
+   * Retrieves the tile min value for a given attribute or dimension and tile
+   * index.
+   *
+   * @param name The input attribute/dimension.
+   * @param tile_idx The index of the tile in the metadata.
+   * @return Status, value, size.
+   */
+  tuple<Status, optional<void*>, optional<uint64_t>> get_tile_min(
+      const std::string& name, uint64_t tile_idx);
+
+  /**
+   * Retrieves the tile max value for a given attribute or dimension and tile
+   * index.
+   *
+   * @param name The input attribute/dimension.
+   * @param tile_idx The index of the tile in the metadata.
+   * @return Status, value, size.
+   */
+  tuple<Status, optional<void*>, optional<uint64_t>> get_tile_max(
+      const std::string& name, uint64_t tile_idx);
+
+  /**
+   * Retrieves the tile sum value for a given attribute or dimension and tile
+   * index.
+   *
+   * @param name The input attribute/dimension.
+   * @param tile_idx The index of the tile in the metadata.
+   * @return Status, sum.
+   */
+  tuple<Status, optional<void*>> get_tile_sum(
+      const std::string& name, uint64_t tile_idx);
+
+  /**
+   * Retrieves the tile null count value for a given attribute or dimension
+   * and tile index.
+   *
+   * @param name The input attribute/dimension.
+   * @param tile_idx The index of the tile in the metadata.
+   * @return Status, count.
+   */
+  tuple<Status, optional<uint64_t>> get_tile_null_count(
+      const std::string& name, uint64_t tile_idx);
+
+  /**
+   * Retrieves the min value for a given attribute or dimension.
+   *
+   * @param name The input attribute/dimension.
+   * @return Status, value.
+   */
+  tuple<Status, optional<std::vector<uint8_t>>> get_min(
+      const std::string& name);
+
+  /**
+   * Retrieves the max value for a given attribute or dimension.
+   *
+   * @param name The input attribute/dimension.
+   * @return Status, value.
+   */
+  tuple<Status, optional<std::vector<uint8_t>>> get_max(
+      const std::string& name);
+
+  /**
+   * Retrieves the sum value for a given attribute or dimension.
+   *
+   * @param name The input attribute/dimension.
+   * @return Status, sum.
+   */
+  tuple<Status, optional<void*>> get_sum(const std::string& name);
+
+  /**
+   * Retrieves the null count value for a given attribute or dimension.
+   *
+   * @param name The input attribute/dimension.
+   * @return Status, count.
+   */
+  tuple<Status, optional<uint64_t>> get_null_count(const std::string& name);
 
   /** Returns the first timestamp of the fragment timestamp range. */
   uint64_t first_timestamp() const;
@@ -517,21 +723,60 @@ class FragmentMetadata {
       const EncryptionKey& encryption_key, std::vector<std::string>&& names);
 
   /**
-   * Loads validity tile offsets for the attribute names.
+   * Loads min values for the attribute names.
    *
    * @param encryption_key The key the array got opened with.
    * @param names The attribute names.
    * @return Status
    */
-  Status load_tile_validity_offsets(
+  Status load_tile_min_values(
       const EncryptionKey& encryption_key, std::vector<std::string>&& names);
 
+  /**
+   * Loads max values for the attribute names.
+   *
+   * @param encryption_key The key the array got opened with.
+   * @param names The attribute names.
+   * @return Status
+   */
+  Status load_tile_max_values(
+      const EncryptionKey& encryption_key, std::vector<std::string>&& names);
+
+  /**
+   * Loads sum values for the attribute names.
+   *
+   * @param encryption_key The key the array got opened with.
+   * @param names The attribute names.
+   * @return Status
+   */
+  Status load_tile_sum_values(
+      const EncryptionKey& encryption_key, std::vector<std::string>&& names);
+
+  /**
+   * Loads null count values for the attribute names.
+   *
+   * @param encryption_key The key the array got opened with.
+   * @param names The attribute names.
+   * @return Status
+   */
+  Status load_tile_null_count_values(
+      const EncryptionKey& encryption_key, std::vector<std::string>&& names);
+
+  /**
+   * Loads the min max sum null count values for the fragment.
+   *
+   * @param encryption_key The key the array got opened with.
+   * @return Status
+   */
+  Status load_fragment_min_max_sum_null_count(
+      const EncryptionKey& encryption_key);
+
   /**
    * Returns ArraySchema
    *
    * @return
    */
-  const ArraySchema* array_schema() const;
+  const shared_ptr<const ArraySchema>& array_schema() const;
 
  private:
   /* ********************************* */
@@ -549,6 +794,11 @@ class FragmentMetadata {
     std::vector<uint64_t> tile_var_offsets_;
     std::vector<uint64_t> tile_var_sizes_;
     std::vector<uint64_t> tile_validity_offsets_;
+    std::vector<uint64_t> tile_min_offsets_;
+    std::vector<uint64_t> tile_max_offsets_;
+    std::vector<uint64_t> tile_sum_offsets_;
+    std::vector<uint64_t> tile_null_count_offsets_;
+    uint64_t tile_min_max_sum_null_count_offset_;
   };
 
   /** Keeps track of which metadata is loaded. */
@@ -559,6 +809,11 @@ class FragmentMetadata {
     std::vector<bool> tile_var_offsets_;
     std::vector<bool> tile_var_sizes_;
     std::vector<bool> tile_validity_offsets_;
+    std::vector<bool> tile_min_;
+    std::vector<bool> tile_max_;
+    std::vector<bool> tile_sum_;
+    std::vector<bool> tile_null_count_;
+    bool fragment_min_max_sum_null_count_ = false;
   };
 
   /* ********************************* */
@@ -568,8 +823,13 @@ class FragmentMetadata {
   /** The storage manager. */
   StorageManager* storage_manager_;
 
+  /**
+   * The memory tracker of the array this fragment metadata corresponds to.
+   */
+  MemoryTracker* memory_tracker_;
+
   /** The array schema */
-  const ArraySchema* array_schema_;
+  shared_ptr<const ArraySchema> array_schema_;
 
   /** The array schema name */
   std::string array_schema_name_;
@@ -673,6 +933,58 @@ class FragmentMetadata {
    */
   std::vector<std::vector<uint64_t>> tile_validity_offsets_;
 
+  /**
+   * The tile min buffers, for variable attributes/dimensions, this will store
+   * offsets.
+   */
+  std::vector<std::vector<uint8_t>> tile_min_buffer_;
+
+  /**
+   * The tile min buffers variable length data.
+   */
+  std::vector<std::vector<char>> tile_min_var_buffer_;
+
+  /**
+   * The tile max buffers, for variable attributes/dimensions, this will store
+   * offsets.
+   */
+  std::vector<std::vector<uint8_t>> tile_max_buffer_;
+
+  /**
+   * The tile max buffers variable length data.
+   */
+  std::vector<std::vector<char>> tile_max_var_buffer_;
+
+  /**
+   * The tile sum values, ignored for var sized attributes/dimensions.
+   */
+  std::vector<std::vector<uint8_t>> tile_sums_;
+
+  /**
+   * The tile null count values for attributes/dimensions.
+   */
+  std::vector<std::vector<uint64_t>> tile_null_counts_;
+
+  /**
+   * Fragment min values.
+   */
+  std::vector<std::vector<uint8_t>> fragment_mins_;
+
+  /**
+   * Fragment max values.
+   */
+  std::vector<std::vector<uint8_t>> fragment_maxs_;
+
+  /**
+   * Fragment sum values, ignored for var sized attributes/dimensions.
+   */
+  std::vector<uint64_t> fragment_sums_;
+
+  /**
+   * Null count for fragment for attributes/dimensions.
+   */
+  std::vector<uint64_t> fragment_null_counts_;
+
   /** The format version of this metadata. */
   uint32_t version_;
 
@@ -715,9 +1027,17 @@ class FragmentMetadata {
    * Returns the size of the fragment metadata footer
    * (which contains the generic tile offsets) along with its size.
    *
-   * Applicable to format version 7 or higher.
+   * Applicable to format version 7 to 10.
    */
-  uint64_t footer_size_v7_or_higher() const;
+  uint64_t footer_size_v7_v10() const;
+
+  /**
+   * Returns the size of the fragment metadata footer
+   * (which contains the generic tile offsets) along with its size.
+   *
+   * Applicable to format version 11 or higher.
+   */
+  uint64_t footer_size_v11_or_higher() const;
 
   /**
    * Returns the ids (positions) of the tiles overlapping `subarray`.
@@ -763,18 +1083,41 @@ class FragmentMetadata {
   Status load_tile_var_sizes(const EncryptionKey& encryption_key, unsigned idx);
 
   /**
-   * Loads the validity tile offsets for the input attribute idx
-   * from storage.
+   * Loads the validity tile offsets for the input attribute idx from storage.
    */
   Status load_tile_validity_offsets(
       const EncryptionKey& encryption_key, unsigned idx);
 
+  /**
+   * Loads the min values for the input attribute idx from storage.
+   */
+  Status load_tile_min_values(
+      const EncryptionKey& encryption_key, unsigned idx);
+
+  /**
+   * Loads the max values for the input attribute idx from storage.
+   */
+  Status load_tile_max_values(
+      const EncryptionKey& encryption_key, unsigned idx);
+
+  /**
+   * Loads the sum values for the input attribute idx from storage.
+   */
+  Status load_tile_sum_values(
+      const EncryptionKey& encryption_key, unsigned idx);
+
+  /**
+   * Loads the null count values for the input attribute idx from storage.
+   */
+  Status load_tile_null_count_values(
+      const EncryptionKey& encryption_key, unsigned idx);
+
   /** Loads the generic tile offsets from the buffer. */
   Status load_generic_tile_offsets(ConstBuffer* buff);
 
   /**
    * Loads the generic tile offsets from the buffer. Applicable to
-   * versions 4 and 5.
+   * versions 3 and 4.
    */
   Status load_generic_tile_offsets_v3_v4(ConstBuffer* buff);
 
@@ -786,9 +1129,21 @@ class FragmentMetadata {
 
   /**
    * Loads the generic tile offsets from the buffer. Applicable to
-   * versions 7 or higher.
+   * versions 7 to 10.
    */
-  Status load_generic_tile_offsets_v7_or_higher(ConstBuffer* buff);
+  Status load_generic_tile_offsets_v7_v10(ConstBuffer* buff);
+
+  /**
+   * Loads the generic tile offsets from the buffer. Applicable to
+   * versions 11.
+   */
+  Status load_generic_tile_offsets_v11(ConstBuffer* buff);
+
+  /**
+   * Loads the generic tile offsets from the buffer. Applicable to
+   * versions 12 or higher.
+   */
+  Status load_generic_tile_offsets_v12_or_higher(ConstBuffer* buff);
 
   /**
    * Loads the array schema name.
@@ -915,6 +1270,32 @@ class FragmentMetadata {
    */
   Status load_tile_validity_offsets(unsigned idx, ConstBuffer* buff);
 
+  /**
+   * Loads the min values for the input attribute from the input buffer.
+   */
+  Status load_tile_min_values(unsigned idx, ConstBuffer* buff);
+
+  /**
+   * Loads the max values for the input attribute from the input buffer.
+   */
+  Status load_tile_max_values(unsigned idx, ConstBuffer* buff);
+
+  /**
+   * Loads the sum values for the input attribute from the input buffer.
+   */
+  Status load_tile_sum_values(unsigned idx, ConstBuffer* buff);
+
+  /**
+   * Loads the null count values for the input attribute from the input
+   * buffer.
+   */
+  Status load_tile_null_count_values(unsigned idx, ConstBuffer* buff);
+
+  /**
+   * Loads the min max sum null count values for the fragment.
+   */
+  Status load_fragment_min_max_sum_null_count(ConstBuffer* buff);
+
   /** Loads the format version from the buffer. */
   Status load_version(ConstBuffer* buff);
 
@@ -927,7 +1308,7 @@ class FragmentMetadata {
   /** Loads the basic metadata from storage (version 2 or before). */
   Status load_v1_v2(
       const EncryptionKey& encryption_key,
-      const std::unordered_map<std::string, tiledb_shared_ptr<ArraySchema>>&
+      const std::unordered_map<std::string, shared_ptr<ArraySchema>>&
           array_schemas);
 
   /**
@@ -938,8 +1319,7 @@ class FragmentMetadata {
       const EncryptionKey& encryption_key,
       Buffer* f_buff,
       uint64_t offset,
-      std::unordered_map<std::string, tiledb_shared_ptr<ArraySchema>>
-          array_schemas);
+      std::unordered_map<std::string, shared_ptr<ArraySchema>> array_schemas);
 
   /**
    * Loads the footer of the metadata file, which contains
@@ -951,8 +1331,7 @@ class FragmentMetadata {
       const EncryptionKey& encryption_key,
       Buffer* f_buff,
       uint64_t offset,
-      std::unordered_map<std::string, tiledb_shared_ptr<ArraySchema>>
-          array_schemas);
+      std::unordered_map<std::string, shared_ptr<ArraySchema>> array_schemas);
 
   /** Writes the sizes of each attribute file to the buffer. */
   Status write_file_sizes(Buffer* buff) const;
@@ -1063,6 +1442,105 @@ class FragmentMetadata {
    */
   Status write_tile_validity_offsets(unsigned idx, Buffer* buff);
 
+  /**
+   * Writes the mins of the input attribute to storage.
+   *
+   * @param idx The index of the attribute.
+   * @param encryption_key The encryption key.
+   * @param nbytes The total number of bytes written for the mins.
+   * @return Status
+   */
+  Status store_tile_mins(
+      unsigned idx, const EncryptionKey& encryption_key, uint64_t* nbytes);
+
+  /**
+   * Writes the mins of the input attribute idx to the input buffer.
+   */
+  Status write_tile_mins(unsigned idx, Buffer* buff);
+
+  /**
+   * Writes the maxs of the input attribute to storage.
+   *
+   * @param idx The index of the attribute.
+   * @param encryption_key The encryption key.
+   * @param nbytes The total number of bytes written for the maxs.
+   * @return Status
+   */
+  Status store_tile_maxs(
+      unsigned idx, const EncryptionKey& encryption_key, uint64_t* nbytes);
+
+  /**
+   * Writes the maxs of the input attribute idx to the input buffer.
+   */
+  Status write_tile_maxs(unsigned idx, Buffer* buff);
+
+  /**
+   * Writes the sums of the input attribute to storage.
+   *
+   * @param idx The index of the attribute.
+   * @param encryption_key The encryption key.
+   * @param nbytes The total number of bytes written for the sums.
+   * @return Status
+   */
+  Status store_tile_sums(
+      unsigned idx, const EncryptionKey& encryption_key, uint64_t* nbytes);
+
+  /**
+   * Writes the sums of the input attribute idx to the input buffer.
+   */
+  Status write_tile_sums(unsigned idx, Buffer* buff);
+
+  /**
+   * Writes the null counts of the input attribute to storage.
+   *
+   * @param idx The index of the attribute.
+   * @param encryption_key The encryption key.
+   * @param nbytes The total number of bytes written for the null counts.
+   * @return Status
+   */
+  Status store_tile_null_counts(
+      unsigned idx, const EncryptionKey& encryption_key, uint64_t* nbytes);
+
+  /**
+   * Writes the null counts of the input attribute idx to the input buffer.
+   */
+  Status write_tile_null_counts(unsigned idx, Buffer* buff);
+
+  /**
+   * Writes the fragment min, max, sum and null count to storage.
+   *
+   * @param num The number of attributes.
+   * @param encryption_key The encryption key.
+   * @param nbytes The total number of bytes written.
+   * @return Status
+   */
+  Status store_fragment_min_max_sum_null_count(
+      uint64_t num, const EncryptionKey& encryption_key, uint64_t* nbytes);
+
+  /**
+   * Compute the fragment min, max and sum values.
+   *
+   * @param name The attribute/dimension name.
+   */
+  template <class T>
+  void compute_fragment_min_max_sum(const std::string& name);
+
+  /**
+   * Compute the fragment sum value.
+   *
+   * @param idx The attribute/dimension index.
+   * @param nullable Is the attribute/dimension nullable.
+   */
+  template <class SumT>
+  void compute_fragment_sum(const uint64_t idx, const bool nullable);
+
+  /**
+   * Compute the fragment min and max values for var sized attributes.
+   *
+   * @param name The attribute/dimension name.
+   */
+  void min_max_var(const std::string& name);
+
   /** Writes the format version to the buffer. */
   Status write_version(Buffer* buff) const;
 
@@ -1097,7 +1575,7 @@ class FragmentMetadata {
    */
   Status write_generic_tile_to_file(
       const EncryptionKey& encryption_key,
-      Buffer&& buff,
+      Buffer& buff,
       uint64_t* nbytes) const;
 
   /**
@@ -1119,9 +1597,10 @@ class FragmentMetadata {
    * motiviation is to encode illegal/reserved file name characters.
    *
    * @param name The dimension/attribute name.
-   * return std::string The encoded dimension/attribute name.
+   * return Status, the encoded dimension/attribute name.
    */
-  std::string encode_name(const std::string& name) const;
+  tuple<Status, optional<std::string>> encode_name(
+      const std::string& name) const;
 
   /**
    * This builds the index mapping for attribute/dimension name to id.
