@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2021 TileDB, Inc.
+ * @copyright Copyright (c) 2021-2022 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -35,9 +35,11 @@
 #include "tiledb/sm/array/array.h"
 #include "tiledb/sm/array_schema/dimension.h"
 #include "tiledb/sm/array_schema/domain.h"
+#include "tiledb/sm/enums/layout.h"
 #include "tiledb/sm/misc/constants.h"
 #include "tiledb/sm/misc/parallel_functions.h"
-#include "tiledb/sm/tile/tile.h"
+#include "tiledb/sm/misc/utils.h"
+#include "tiledb/sm/tile/writer_tile.h"
 
 using namespace tiledb::common;
 using namespace tiledb::sm::stats;
@@ -68,7 +70,7 @@ DenseTiler<T>::DenseTiler(
   assert(subarray != nullptr);
   assert(buffers != nullptr);
   for (const auto& buff : *buffers) {
-    assert(array_schema_->is_attr(buff.first));
+    assert(array_schema_.is_attr(buff.first));
     (void)buff;
   }
 
@@ -94,13 +96,13 @@ const typename DenseTiler<T>::CopyPlan DenseTiler<T>::copy_plan(
 
   // For easy reference
   CopyPlan ret;
-  auto dim_num = (int32_t)array_schema_->dim_num();
-  auto domain = array_schema_->domain();
+  auto dim_num = (int32_t)array_schema_.dim_num();
+  auto& domain{array_schema_.domain()};
   auto subarray = subarray_->ndrange(0);  // Guaranteed to be unary
   std::vector<std::array<T, 2>> sub(dim_num);
   for (int32_t d = 0; d < dim_num; ++d)
     sub[d] = {*(const T*)subarray[d].start(), *(const T*)subarray[d].end()};
-  auto tile_layout = array_schema_->cell_order();
+  auto tile_layout = array_schema_.cell_order();
   auto sub_layout = subarray_->layout();
 
   // Copy tile and subarray strides
@@ -140,7 +142,7 @@ const typename DenseTiler<T>::CopyPlan DenseTiler<T>::copy_plan(
           sub_in_tile[dim_num - 1][1] - sub_in_tile[dim_num - 1][0] + 1;
       int32_t last_d = dim_num - 2;
       for (; last_d >= 0; --last_d) {
-        auto tile_extent = *(const T*)domain->tile_extent(last_d + 1).data();
+        auto tile_extent = *(const T*)domain.tile_extent(last_d + 1).data();
         if (sub_in_tile[last_d + 1][1] - sub_in_tile[last_d + 1][0] + 1 ==
                 tile_extent &&
             sub_in_tile[last_d + 1][0] == sub[last_d + 1][0] &&
@@ -161,7 +163,7 @@ const typename DenseTiler<T>::CopyPlan DenseTiler<T>::copy_plan(
       ret.copy_el_ = sub_in_tile[0][1] - sub_in_tile[0][0] + 1;
       int32_t last_d = 1;
       for (; last_d < dim_num; ++last_d) {
-        auto tile_extent = *(const T*)domain->tile_extent(last_d - 1).data();
+        auto tile_extent = *(const T*)domain.tile_extent(last_d - 1).data();
         if (sub_in_tile[last_d - 1][1] - sub_in_tile[last_d - 1][0] + 1 ==
                 tile_extent &&
             sub_in_tile[last_d - 1][0] == sub[last_d - 1][0] &&
@@ -187,26 +189,26 @@ const typename DenseTiler<T>::CopyPlan DenseTiler<T>::copy_plan(
 
 template <class T>
 Status DenseTiler<T>::get_tile(
-    uint64_t id, const std::string& name, Tile* tile) {
+    uint64_t id, const std::string& name, WriterTile* tile) {
   auto timer_se = stats_->start_timer("get_tile");
 
   // Checks
   if (id >= tile_num_)
     return LOG_STATUS(
-        Status::DenseTilerError("Cannot get tile; Invalid tile id"));
-  if (!array_schema_->is_attr(name))
-    return LOG_STATUS(Status::DenseTilerError(
+        Status_DenseTilerError("Cannot get tile; Invalid tile id"));
+  if (!array_schema_.is_attr(name))
+    return LOG_STATUS(Status_DenseTilerError(
         std::string("Cannot get tile; '") + name + "' is not an attribute"));
-  if (array_schema_->var_size(name))
-    return LOG_STATUS(Status::DenseTilerError(
+  if (array_schema_.var_size(name))
+    return LOG_STATUS(Status_DenseTilerError(
         std::string("Cannot get tile; '") + name +
         "' is not a fixed-sized attribute"));
 
   // For easy reference
-  auto domain = array_schema_->domain();
-  auto type = array_schema_->type(name);
-  auto cell_num_per_tile = domain->cell_num_per_tile();
-  auto cell_size = array_schema_->cell_size(name);
+  auto& domain{array_schema_.domain()};
+  auto type = array_schema_.type(name);
+  auto cell_num_per_tile = domain.cell_num_per_tile();
+  auto cell_size = array_schema_.cell_size(name);
   auto tile_size = cell_num_per_tile * cell_size;
   auto buff = (uint8_t*)buffers_->find(name)->second.buffer_;
   assert(buff != nullptr);
@@ -214,7 +216,6 @@ Status DenseTiler<T>::get_tile(
   // Initialize tile
   RETURN_NOT_OK(tile->init_unfiltered(
       constants::format_version, type, tile_size, cell_size, 0, true));
-  tile->reset_offset();
 
   // Copy tile from buffer
   RETURN_NOT_OK(copy_tile(id, cell_size, buff, tile));
@@ -224,23 +225,23 @@ Status DenseTiler<T>::get_tile(
 
 template <class T>
 Status DenseTiler<T>::get_tile_null(
-    uint64_t id, const std::string& name, Tile* tile) const {
+    uint64_t id, const std::string& name, WriterTile* tile) const {
   // Checks
   if (id >= tile_num_)
     return LOG_STATUS(
-        Status::DenseTilerError("Cannot get tile; Invalid tile id"));
-  if (!array_schema_->is_attr(name))
-    return LOG_STATUS(Status::DenseTilerError(
+        Status_DenseTilerError("Cannot get tile; Invalid tile id"));
+  if (!array_schema_.is_attr(name))
+    return LOG_STATUS(Status_DenseTilerError(
         std::string("Cannot get tile; '") + name + "' is not an attribute"));
-  if (!array_schema_->is_nullable(name))
-    return LOG_STATUS(Status::DenseTilerError(
+  if (!array_schema_.is_nullable(name))
+    return LOG_STATUS(Status_DenseTilerError(
         std::string("Cannot get tile; '") + name +
         "' is not a nullable attribute"));
 
   // For easy reference
-  auto domain = array_schema_->domain();
-  auto type = array_schema_->type(name);
-  auto cell_num_per_tile = domain->cell_num_per_tile();
+  auto& domain{array_schema_.domain()};
+  auto type = array_schema_.type(name);
+  auto cell_num_per_tile = domain.cell_num_per_tile();
   auto cell_size = constants::cell_validity_size;
   auto tile_size = cell_num_per_tile * cell_size;
   auto buff = (uint8_t*)buffers_->find(name)->second.validity_vector_.buffer();
@@ -249,8 +250,6 @@ Status DenseTiler<T>::get_tile_null(
   // Initialize tile
   RETURN_NOT_OK(tile->init_unfiltered(
       constants::format_version, type, tile_size, cell_size, 0, true));
-  tile->reset_offset();
-  assert(tile->size() == tile_size);
 
   // Copy tile from buffer
   return copy_tile(id, cell_size, buff, tile);
@@ -260,24 +259,24 @@ template <class T>
 Status DenseTiler<T>::get_tile_var(
     uint64_t id,
     const std::string& name,
-    Tile* tile_off,
-    Tile* tile_val) const {
+    WriterTile* tile_off,
+    WriterTile* tile_val) const {
   // Checks
   if (id >= tile_num_)
     return LOG_STATUS(
-        Status::DenseTilerError("Cannot get tile; Invalid tile id"));
-  if (!array_schema_->is_attr(name))
-    return LOG_STATUS(Status::DenseTilerError(
+        Status_DenseTilerError("Cannot get tile; Invalid tile id"));
+  if (!array_schema_.is_attr(name))
+    return LOG_STATUS(Status_DenseTilerError(
         std::string("Cannot get tile; '") + name + "' is not an attribute"));
-  if (!array_schema_->var_size(name))
-    return LOG_STATUS(Status::DenseTilerError(
+  if (!array_schema_.var_size(name))
+    return LOG_STATUS(Status_DenseTilerError(
         std::string("Cannot get tile; '") + name +
         "' is not a var-sized attribute"));
 
   // For easy reference
-  auto domain = array_schema_->domain();
-  auto type = array_schema_->type(name);
-  auto cell_num_in_tile = domain->cell_num_per_tile();
+  auto& domain{array_schema_.domain()};
+  auto type = array_schema_.type(name);
+  auto cell_num_in_tile = domain.cell_num_per_tile();
   auto tile_off_size = constants::cell_var_offset_size * cell_num_in_tile;
   auto buff_it = buffers_->find(name);
   assert(buff_it != buffers_->end());
@@ -289,7 +288,7 @@ Status DenseTiler<T>::get_tile_var(
   std::vector<uint8_t> fill_var(sizeof(uint64_t), 0);
 
   // Initialize position tile
-  Tile tile_pos;
+  WriterTile tile_pos;
   RETURN_NOT_OK(tile_pos.init_unfiltered(
       constants::format_version,
       constants::cell_var_offset_type,
@@ -300,10 +299,9 @@ Status DenseTiler<T>::get_tile_var(
   // Fill entire tile with MAX_UINT64
   std::vector<uint64_t> to_write(
       cell_num_in_tile, std::numeric_limits<uint64_t>::max());
-  tile_pos.write(to_write.data(), tile_off_size);
+  RETURN_NOT_OK(tile_pos.write(to_write.data(), 0, tile_off_size));
   to_write.clear();
   to_write.shrink_to_fit();
-  tile_pos.reset_offset();
 
   // Get position tile
   auto cell_num_in_buff =  // TODO: fix
@@ -330,15 +328,16 @@ Status DenseTiler<T>::get_tile_var(
       0));
 
   // Copy real offsets and values to the corresponding tiles
-  void* tile_pos_buff_tmp = tile_pos.buffer()->data();
+  void* tile_pos_buff_tmp = tile_pos.data();
   auto tile_pos_buff = (uint64_t*)tile_pos_buff_tmp;
-  uint64_t offset = 0, val_offset, val_size, pos;
+  uint64_t tile_off_offset = 0, offset = 0, val_offset, val_size, pos;
   auto mul = (offsets_format_mode_ == "bytes") ? 1 : cell_size;
   for (uint64_t i = 0; i < cell_num_in_tile; ++i) {
     pos = tile_pos_buff[i];
-    tile_off->write(&offset, sizeof(offset));
+    RETURN_NOT_OK(tile_off->write(&offset, tile_off_offset, sizeof(offset)));
+    tile_off_offset += sizeof(offset);
     if (pos == std::numeric_limits<uint64_t>::max()) {  // Empty
-      tile_val->write(&fill_var[0], cell_size);
+      RETURN_NOT_OK(tile_val->write_var(&fill_var[0], offset, cell_size));
       offset += cell_size;
     } else {  // Non-empty
       val_offset = ((offsets_bytesize_ == 8) ?
@@ -352,14 +351,13 @@ Status DenseTiler<T>::get_tile_var(
                           mul -
                       val_offset) :
                      buff_var_size - val_offset;
-      tile_val->write(&buff_var[val_offset], val_size);
+      RETURN_NOT_OK(
+          tile_val->write_var(&buff_var[val_offset], offset, val_size));
       offset += val_size;
     }
   }
 
-  // Reset tile offsets
-  tile_off->reset_offset();
-  tile_val->reset_offset();
+  tile_val->final_size(offset);
 
   return Status::Ok();
 }
@@ -396,8 +394,8 @@ const std::vector<uint64_t>& DenseTiler<T>::first_sub_tile_coords() const {
 template <class T>
 void DenseTiler<T>::calculate_first_sub_tile_coords() {
   // For easy reference
-  auto dim_num = array_schema_->dim_num();
-  auto domain = array_schema_->domain();
+  auto dim_num = array_schema_.dim_num();
+  auto& domain{array_schema_.domain()};
   auto subarray = subarray_->ndrange(0);
 
   // Calculate the coordinates of the first tile in the entire
@@ -405,9 +403,9 @@ void DenseTiler<T>::calculate_first_sub_tile_coords() {
   // left cell)
   first_sub_tile_coords_.resize(dim_num);
   for (unsigned d = 0; d < dim_num; ++d) {
-    T dom_start = *(const T*)domain->dimension(d)->domain().start();
+    T dom_start = *(const T*)domain.dimension(d)->domain().start();
     T sub_start = *(const T*)subarray[d].start();
-    T tile_extent = *(const T*)domain->tile_extent(d).data();
+    T tile_extent = *(const T*)domain.tile_extent(d).data();
     first_sub_tile_coords_[d] =
         Dimension::tile_idx(sub_start, dom_start, tile_extent);
   }
@@ -416,17 +414,17 @@ void DenseTiler<T>::calculate_first_sub_tile_coords() {
 template <class T>
 void DenseTiler<T>::calculate_subarray_tile_coord_strides() {
   // For easy reference
-  auto dim_num = (int32_t)array_schema_->dim_num();
-  auto domain = array_schema_->domain();
+  auto dim_num = (int32_t)array_schema_.dim_num();
+  auto& domain{array_schema_.domain()};
   auto subarray = subarray_->ndrange(0);
-  auto layout = array_schema_->tile_order();
+  auto layout = array_schema_.tile_order();
 
   // Compute offsets
   sub_tile_coord_strides_.reserve(dim_num);
   if (layout == Layout::ROW_MAJOR) {
     sub_tile_coord_strides_.push_back(1);
     for (auto d = dim_num - 2; d >= 0; --d) {
-      auto tile_num = domain->dimension(d + 1)->tile_num(subarray[d + 1]);
+      auto tile_num = domain.dimension(d + 1)->tile_num(subarray[d + 1]);
       sub_tile_coord_strides_.push_back(
           sub_tile_coord_strides_.back() * tile_num);
     }
@@ -435,7 +433,7 @@ void DenseTiler<T>::calculate_subarray_tile_coord_strides() {
   } else {  // COL_MAJOR
     sub_tile_coord_strides_.push_back(1);
     for (int32_t d = 1; d < dim_num; ++d) {
-      auto tile_num = domain->dimension(d - 1)->tile_num(subarray[d - 1]);
+      auto tile_num = domain.dimension(d - 1)->tile_num(subarray[d - 1]);
       sub_tile_coord_strides_.push_back(
           sub_tile_coord_strides_.back() * tile_num);
     }
@@ -447,9 +445,9 @@ void DenseTiler<T>::calculate_tile_and_subarray_strides() {
   // For easy reference
   auto sub_layout = subarray_->layout();
   assert(sub_layout == Layout::ROW_MAJOR || sub_layout == Layout::COL_MAJOR);
-  auto tile_layout = array_schema_->cell_order();
-  auto dim_num = (int32_t)array_schema_->dim_num();
-  auto domain = array_schema_->domain();
+  auto tile_layout = array_schema_.cell_order();
+  auto dim_num = (int32_t)array_schema_.dim_num();
+  auto& domain{array_schema_.domain()};
   auto subarray = subarray_->ndrange(0);
 
   // Compute tile strides
@@ -458,7 +456,7 @@ void DenseTiler<T>::calculate_tile_and_subarray_strides() {
     tile_strides_el_[dim_num - 1] = 1;
     if (dim_num > 1) {
       for (auto d = dim_num - 2; d >= 0; --d) {
-        auto tile_extent = (const T*)(domain->tile_extent(d + 1).data());
+        auto tile_extent = (const T*)(domain.tile_extent(d + 1).data());
         assert(tile_extent != nullptr);
         tile_strides_el_[d] = Dimension::tile_extent_mult<T>(
             tile_strides_el_[d + 1], *tile_extent);
@@ -468,7 +466,7 @@ void DenseTiler<T>::calculate_tile_and_subarray_strides() {
     tile_strides_el_[0] = 1;
     if (dim_num > 1) {
       for (auto d = 1; d < dim_num; ++d) {
-        auto tile_extent = (const T*)(domain->tile_extent(d - 1).data());
+        auto tile_extent = (const T*)(domain.tile_extent(d - 1).data());
         assert(tile_extent != nullptr);
         tile_strides_el_[d] = Dimension::tile_extent_mult<T>(
             tile_strides_el_[d - 1], *tile_extent);
@@ -503,14 +501,14 @@ void DenseTiler<T>::calculate_tile_and_subarray_strides() {
 
 template <class T>
 void DenseTiler<T>::calculate_tile_num() {
-  tile_num_ = array_schema_->domain()->tile_num(subarray_->ndrange(0));
+  tile_num_ = array_schema_.domain().tile_num(subarray_->ndrange(0));
 }
 
 template <class T>
 std::vector<uint64_t> DenseTiler<T>::tile_coords_in_sub(uint64_t id) const {
   // For easy reference
-  auto dim_num = (int32_t)array_schema_->dim_num();
-  auto layout = array_schema_->tile_order();
+  auto dim_num = (int32_t)array_schema_.dim_num();
+  auto layout = array_schema_.tile_order();
   std::vector<uint64_t> ret;
   auto tmp_idx = id;
 
@@ -533,8 +531,8 @@ std::vector<uint64_t> DenseTiler<T>::tile_coords_in_sub(uint64_t id) const {
 template <class T>
 std::vector<std::array<T, 2>> DenseTiler<T>::tile_subarray(uint64_t id) const {
   // For easy reference
-  auto dim_num = array_schema_->dim_num();
-  auto domain = array_schema_->domain();
+  auto dim_num = array_schema_.dim_num();
+  auto& domain{array_schema_.domain()};
 
   // Get tile coordinates in the subarray tile domain
   auto tile_coords_in_sub = this->tile_coords_in_sub(id);
@@ -547,8 +545,8 @@ std::vector<std::array<T, 2>> DenseTiler<T>::tile_subarray(uint64_t id) const {
   // Calculate tile subarray based on the tile coordinates in the domain
   std::vector<std::array<T, 2>> ret(dim_num);
   for (unsigned d = 0; d < dim_num; ++d) {
-    auto dom_start = *(const T*)domain->dimension(d)->domain().start();
-    auto tile_extent = *(const T*)domain->tile_extent(d).data();
+    auto dom_start = *(const T*)domain.dimension(d)->domain().start();
+    auto tile_extent = *(const T*)domain.tile_extent(d).data();
     ret[d][0] = Dimension::tile_coord_low(
         tile_coords_in_dom[d], dom_start, tile_extent);
     ret[d][1] = Dimension::tile_coord_high(
@@ -560,7 +558,7 @@ std::vector<std::array<T, 2>> DenseTiler<T>::tile_subarray(uint64_t id) const {
 
 template <class T>
 Status DenseTiler<T>::copy_tile(
-    uint64_t id, uint64_t cell_size, uint8_t* buff, Tile* tile) const {
+    uint64_t id, uint64_t cell_size, uint8_t* buff, WriterTile* tile) const {
   // Calculate copy plan
   const CopyPlan copy_plan = this->copy_plan(id);
 
@@ -622,9 +620,6 @@ Status DenseTiler<T>::copy_tile(
     }
   }
 
-  // Reset the tile offset to the beginning of the tile
-  tile->reset_offset();
-
   return Status::Ok();
 }
 
