@@ -366,7 +366,7 @@ tuple<Status, optional<bool>> SparseGlobalOrderReader::create_result_tiles() {
           auto tile_num = fragment_metadata_[f]->tile_num();
 
           // Figure out the start index.
-          auto start = read_state_.frag_tile_idx_[f].first;
+          auto start = read_state_.frag_idx_[f].tile_idx_;
           if (!result_tiles_[f].empty()) {
             start = std::max(start, result_tiles_[f].back().tile_idx() + 1);
           }
@@ -459,65 +459,52 @@ SparseGlobalOrderReader::compute_result_cell_slab() {
   }
 }
 
-template <class T>
-tuple<Status, optional<bool>> SparseGlobalOrderReader::add_next_tile_to_queue(
-    unsigned int frag_idx,
-    uint64_t cell_idx,
-    std::vector<std::list<ResultTileWithBitmap<uint8_t>>::iterator>&
-        result_tiles_it,
-    std::vector<uint8_t>& result_tile_used,
-    std::priority_queue<ResultCoords, std::vector<ResultCoords>, T>& tile_queue,
-    std::mutex& tile_queue_mutex) {
-  bool found = false;
+template <class CompType>
+tuple<Status, optional<bool>> SparseGlobalOrderReader::add_next_cell_to_queue(
+    GlobalOrderResultCoords& rc,
+    std::vector<TileListIt>& result_tiles_it,
+    TileMinHeap<CompType>& tile_queue) {
+  auto frag_idx = rc.tile_->frag_idx();
+
+  // Try the next cell in the same tile.
+  if (rc.advance_to_next_cell()) {
+    // Update the fragment index, add the cell to the queue and return.
+    std::unique_lock<std::mutex> ul(tile_queue_mutex_);
+    read_state_.frag_idx_[frag_idx] = FragIdx(rc.tile_->tile_idx(), rc.pos_);
+    tile_queue.emplace(std::move(rc));
+    return {Status::Ok(), false};
+  }
+
+  // Save the potential tile to delete and increment the tile iterator.
+  auto to_delete = result_tiles_it[frag_idx];
+  result_tiles_it[frag_idx]++;
 
   // Remove the tile from result tiles if it wasn't used at all.
-  if (!result_tile_used[frag_idx]) {
-    auto to_delete = result_tiles_it[frag_idx];
-    to_delete--;
+  if (!rc.tile_->used()) {
     remove_result_tile(frag_idx, to_delete);
   }
 
-  // Try to find a tile.
-  while (!found && result_tiles_it[frag_idx] != result_tiles_[frag_idx].end()) {
-    bool has_bmp = result_tiles_it[frag_idx]->bitmap_.size() > 0;
-    found = !has_bmp;
-    auto tile = &*result_tiles_it[frag_idx];
+  // Try to find a new tile.
+  if (result_tiles_it[frag_idx] != result_tiles_[frag_idx].end()) {
+    // Find a cell in the current result tile.
+    GlobalOrderResultCoords rc(&*result_tiles_it[frag_idx], 0);
 
-    // Find a cell that's in the subarray.
-    if (has_bmp) {
-      while (cell_idx < tile->cell_num()) {
-        if (tile->bitmap_.size() == 0 || tile->bitmap_[cell_idx]) {
-          found = true;
-          break;
-        }
-
-        cell_idx++;
-      }
+    // All tiles should at least have one cell available.
+    if (!rc.advance_to_next_cell()) {
+      throw std::logic_error("All tiles should have at least one cell.");
     }
 
-    // There was more tiles in this fragment, insert it in the queue.
-    if (found) {
-      std::unique_lock<std::mutex> ul(tile_queue_mutex);
-      tile_queue.emplace(tile, cell_idx);
-      result_tiles_it[frag_idx]++;
-    } else {
-      // Remove the tile.
-      auto to_delete = result_tiles_it[frag_idx];
-      result_tiles_it[frag_idx]++;
-      remove_result_tile(frag_idx, to_delete);
+    // Insert the cell in the queue.
+    {
+      std::unique_lock<std::mutex> ul(tile_queue_mutex_);
+      tile_queue.emplace(std::move(rc));
     }
-
-    result_tile_used[frag_idx] = false;
-
-    // Once we move to the next tile, the saved cell index doesn't matter.
-    cell_idx = 0;
-  }
-
-  if (!found) {
+    read_state_.frag_idx_[frag_idx] = FragIdx(rc.tile_->tile_idx(), rc.pos_);
+  } else {
     // Increment the tile index, which should clear all tiles in end_iteration.
     if (!result_tiles_[frag_idx].empty()) {
-      read_state_.frag_tile_idx_[frag_idx].first++;
-      read_state_.frag_tile_idx_[frag_idx].second = 0;
+      read_state_.frag_idx_[frag_idx].tile_idx_++;
+      read_state_.frag_idx_[frag_idx].cell_idx_ = 0;
     }
 
     // This fragment has more tiles potentially.
@@ -527,6 +514,7 @@ tuple<Status, optional<bool>> SparseGlobalOrderReader::add_next_tile_to_queue(
     }
   }
 
+  // We don't need more tiles as a tile was found.
   return {Status::Ok(), false};
 }
 
@@ -545,15 +533,15 @@ Status SparseGlobalOrderReader::compute_hilbert_values(
   // Parallelize on tiles.
   auto status = parallel_for(
       storage_manager_->compute_tp(), 0, result_tiles.size(), [&](uint64_t t) {
-        auto tile = (ResultTileWithBitmap<uint8_t>*)result_tiles[t];
+        auto tile = static_cast<GlobalOrderResultTile*>(result_tiles[t]);
         auto cell_num = tile->cell_num();
-        auto rc = ResultCoords(tile, 0);
+        auto rc = GlobalOrderResultCoords(tile, 0);
         std::vector<uint64_t> coords(dim_num);
 
-        tile->hilbert_values_.resize(cell_num);
+        tile->allocate_hilbert_vector();
         for (rc.pos_ = 0; rc.pos_ < cell_num; rc.pos_++) {
           // Process only values in bitmap.
-          if (tile->bitmap_.size() == 0 || tile->bitmap_[rc.pos_]) {
+          if (!tile->has_bmp() || tile->bitmap_[rc.pos_]) {
             // Compute Hilbert number for all dimensions first.
             for (uint32_t d = 0; d < dim_num; ++d) {
               auto dim = array_schema_.dimension(d);
@@ -562,7 +550,7 @@ Status SparseGlobalOrderReader::compute_hilbert_values(
             }
 
             // Now we are ready to get the final number.
-            tile->hilbert_values_[rc.pos_] = h.coords_to_hilbert(&coords[0]);
+            tile->set_hilbert_value(rc.pos_, h.coords_to_hilbert(&coords[0]));
           }
         }
 
@@ -573,9 +561,10 @@ Status SparseGlobalOrderReader::compute_hilbert_values(
   return Status::Ok();
 }
 
-template <class T>
+template <class CompType>
 tuple<Status, optional<std::vector<ResultCellSlab>>>
-SparseGlobalOrderReader::merge_result_cell_slabs(uint64_t num_cells, T cmp) {
+SparseGlobalOrderReader::merge_result_cell_slabs(
+    uint64_t num_cells, CompType cmp) {
   auto timer_se = stats_->start_timer("merge_result_cell_slabs");
   std::vector<ResultCellSlab> result_cell_slabs;
 
@@ -584,44 +573,34 @@ SparseGlobalOrderReader::merge_result_cell_slabs(uint64_t num_cells, T cmp) {
   // For easy reference.
   auto allows_dups = array_schema_.allows_dups();
 
-  // A tile min heap, contains one ResultCoords per fragment.
-  std::vector<ResultCoords> container;
+  // A tile min heap, contains one GlobalOrderResultCoords per fragment.
+  std::vector<GlobalOrderResultCoords> container;
   container.reserve(result_tiles_.size());
-  std::priority_queue<ResultCoords, std::vector<ResultCoords>, T> tile_queue(
-      cmp, std::move(container));
-
-  std::mutex tile_queue_mutex;
+  TileMinHeap<CompType> tile_queue(cmp, std::move(container));
 
   // If any fragments needs to load more tiles.
   bool need_more_tiles = false;
 
   // Tile iterators, per fragments.
-  std::vector<std::list<ResultTileWithBitmap<uint8_t>>::iterator>
-      result_tiles_it(result_tiles_.size());
+  std::vector<TileListIt> result_tiles_it(result_tiles_.size());
 
-  // Boolean per fragment that keeps track of when a tile is used.
-  std::vector<uint8_t> result_tile_used(result_tiles_.size(), true);
-
-  // For all fragments, get the first tile.
+  // For all fragments, get the first tile in the sorting queue.
   auto status = parallel_for(
       storage_manager_->compute_tp(), 0, result_tiles_.size(), [&](uint64_t f) {
         if (result_tiles_[f].size() > 0) {
           // Initialize the iterator for this fragment.
           result_tiles_it[f] = result_tiles_[f].begin();
 
-          // Get the cell index we were processing.
-          auto cell_idx = read_state_.frag_tile_idx_[f].second;
-
           // Add the tile to the queue.
-          auto&& [st, more_tiles] = add_next_tile_to_queue(
-              f,
-              cell_idx,
-              result_tiles_it,
-              result_tile_used,
-              tile_queue,
-              tile_queue_mutex);
+          GlobalOrderResultCoords rc(
+              &*(result_tiles_it[f]), read_state_.frag_idx_[f].cell_idx_);
+          auto&& [st, more_tiles] =
+              add_next_cell_to_queue(rc, result_tiles_it, tile_queue);
           RETURN_NOT_OK(st);
-          need_more_tiles = *more_tiles;
+
+          if (*more_tiles) {
+            need_more_tiles = true;
+          }
         }
 
         return Status::Ok();
@@ -637,150 +616,67 @@ SparseGlobalOrderReader::merge_result_cell_slabs(uint64_t num_cells, T cmp) {
     while (!tile_queue.empty() && to_process.same_coords(tile_queue.top()) &&
            num_cells > 0) {
       // Potentially the next cell.
-      auto next_tile = tile_queue.top();
+      auto to_process_dup = tile_queue.top();
       tile_queue.pop();
 
       // Take the cell with the highest fagment index.
-      if (to_process.tile_->frag_idx() < next_tile.tile_->frag_idx()) {
-        std::swap(to_process, next_tile);
+      if (to_process.tile_->frag_idx() < to_process_dup.tile_->frag_idx()) {
+        std::swap(to_process, to_process_dup);
       }
 
       // If we allow duplicates, create one slab for all the dups.
+      auto tile = to_process_dup.tile_;
       if (allows_dups) {
-        result_tile_used[next_tile.tile_->frag_idx()] = true;
-        result_cell_slabs.emplace_back(next_tile.tile_, next_tile.pos_, 1);
+        tile->set_used();
+        result_cell_slabs.emplace_back(
+            to_process_dup.tile_, to_process_dup.pos_, 1);
         num_cells--;
-        read_state_.frag_tile_idx_[next_tile.tile_->frag_idx()] =
-            std::pair<uint64_t, uint64_t>(
-                next_tile.tile_->tile_idx(), next_tile.pos_);
       }
 
       // Put the next cell in the queue.
-      if (!next_tile.next()) {
-        // Done with this tile, fetch another.
-        auto&& [st, more_tiles] = add_next_tile_to_queue(
-            next_tile.tile_->frag_idx(),
-            0,
-            result_tiles_it,
-            result_tile_used,
-            tile_queue,
-            tile_queue_mutex);
-        RETURN_NOT_OK_TUPLE(st, nullopt);
-        need_more_tiles = *more_tiles;
-      } else {
-        tile_queue.emplace(std::move(next_tile));
-      }
+      auto&& [st, more_tiles] =
+          add_next_cell_to_queue(to_process_dup, result_tiles_it, tile_queue);
+      RETURN_NOT_OK_TUPLE(st, nullopt);
+      need_more_tiles = *more_tiles;
     }
 
     if (num_cells == 0) {
       break;
     }
 
-    // Get the tile and flag it as used.
-    auto tile = (ResultTileWithBitmap<uint8_t>*)to_process.tile_;
-    auto has_bmp = tile->bitmap_.size() > 0;
-    result_tile_used[tile->frag_idx()] = true;
-
-    // Find how many cells to process using the top of the queue.
-    // Temp result coord used to find the last position.
-    ResultCoords temp_rc = to_process;
-
-    // Check the top of the queue against last possible cell in the current
-    // tile.
-    if (!has_bmp) {
-      temp_rc.pos_ =
-          std::min(tile->cell_num() - 1, to_process.pos_ + num_cells - 1);
-    } else {
-      temp_rc.pos_ =
-          tile->pos_with_given_result_sum(to_process.pos_, num_cells);
-    }
-
-    // If there is more than one fragment and we can't add the whole tile,
-    // find the last possible cell in this tile smaller than the top of the
-    // queue. Otherwise we are adding everything.
-    if (!tile_queue.empty()) {
-      auto& next_tile = tile_queue.top();
-      if (cmp(temp_rc, next_tile)) {
-        // Run a bisection seach on to find the last cell.
-        uint64_t left = to_process.pos_;
-        uint64_t right = temp_rc.pos_;
-        while (left != right - 1) {
-          // Check against mid.
-          temp_rc.pos_ = left + (right - left) / 2;
-
-          if (!cmp(temp_rc, next_tile))
-            left = temp_rc.pos_;
-          else
-            right = temp_rc.pos_;
-        }
-
-        // Left is the last position smaller than the top of the queue.
-        temp_rc.pos_ = left;
-      }
-    }
-
-    // Generate the result cell slabs.
+    // Get data from the result coord.
+    auto tile = to_process.tile_;
     auto start = to_process.pos_;
     const auto tile_idx = tile->tile_idx();
-    auto& frag_idx = read_state_.frag_tile_idx_[tile->frag_idx()];
+    const auto frag_idx = tile->frag_idx();
 
-    // If no bitmap is set, add all cells.
-    if (!has_bmp) {
-      auto length = std::min(temp_rc.pos_ - to_process.pos_ + 1, num_cells);
-      result_cell_slabs.emplace_back(tile, start, length);
-      frag_idx = std::pair<uint64_t, uint64_t>(tile_idx, start + length);
-      num_cells -= length;
+    // Flag the tile as used.
+    to_process.tile_->set_used();
+
+    // Compute the length of the cell slab.
+    uint64_t length = std::numeric_limits<uint64_t>::max();
+    if (tile_queue.empty()) {
+      length = to_process.max_slab_length();
     } else {
-      // Process all cells, when there is a "hole" in the cell contiguity,
-      // push a new cell slab.
-      uint64_t length = 0;
-      for (auto c = to_process.pos_; c <= temp_rc.pos_; c++) {
-        if (!tile->bitmap_[c]) {
-          if (length != 0) {
-            result_cell_slabs.emplace_back(tile, start, length);
-            frag_idx = std::pair<uint64_t, uint64_t>(tile_idx, start + length);
-            num_cells -= length;
-            length = 0;
-          }
-
-          start = c + 1;
-        } else {
-          length++;
-
-          if (length == num_cells)
-            break;
-        }
-      }
-
-      // Add the last cell slab.
-      if (length != 0) {
-        result_cell_slabs.emplace_back(tile, start, length);
-        frag_idx = std::pair<uint64_t, uint64_t>(tile_idx, start + length - 1);
-        num_cells -= length;
-      }
+      length = to_process.max_slab_length(tile_queue.top(), cmp);
     }
 
-    // Update the position in the tile.
-    to_process.pos_ = temp_rc.pos_;
+    // Make sure we don't merge more cells than the buffers.
+    length = std::min(length, num_cells);
+
+    // Generate the result cell slabs.
+    result_cell_slabs.emplace_back(tile, start, length);
+    read_state_.frag_idx_[frag_idx] = FragIdx(tile_idx, start + length);
+    num_cells -= length;
+
+    // Update the position in the result coord.
+    to_process.pos_ += length - 1;
 
     // Put the next cell in the queue.
-    if (!to_process.next()) {
-      // Done with this tile, fetch another.
-      auto&& [st, more_tiles] = add_next_tile_to_queue(
-          tile->frag_idx(),
-          0,
-          result_tiles_it,
-          result_tile_used,
-          tile_queue,
-          tile_queue_mutex);
-      RETURN_NOT_OK_TUPLE(st, nullopt);
-      need_more_tiles = *more_tiles;
-    } else {
-      // Put the next cell on the queue to be resorted.
-      read_state_.frag_tile_idx_[tile->frag_idx()] =
-          std::pair<uint64_t, uint64_t>(tile->tile_idx(), to_process.pos_);
-      tile_queue.emplace(std::move(to_process));
-    }
+    auto&& [st, more_tiles] =
+        add_next_cell_to_queue(to_process, result_tiles_it, tile_queue);
+    RETURN_NOT_OK_TUPLE(st, nullopt);
+    need_more_tiles = *more_tiles;
   }
 
   buffers_full_ = num_cells == 0;
@@ -839,7 +735,8 @@ Status SparseGlobalOrderReader::copy_offsets_tiles(
       [&](uint64_t i, uint64_t range_thread_idx) {
         // For easy reference.
         auto& rcs = result_cell_slabs[i];
-        auto rt = (ResultTileWithBitmap<uint8_t>*)result_cell_slabs[i].tile_;
+        auto rt =
+            static_cast<GlobalOrderResultTile*>(result_cell_slabs[i].tile_);
 
         // Get source buffers.
         const auto tile_tuple = rt->tile_tuple(name);
@@ -998,7 +895,8 @@ Status SparseGlobalOrderReader::copy_fixed_data_tiles(
       [&](uint64_t i, uint64_t range_thread_idx) {
         // For easy reference.
         auto& rcs = result_cell_slabs[i];
-        auto rt = (ResultTileWithBitmap<uint8_t>*)result_cell_slabs[i].tile_;
+        auto rt =
+            static_cast<GlobalOrderResultTile*>(result_cell_slabs[i].tile_);
 
         // Get source buffers.
         const auto stores_zipped_coords = is_dim && rt->stores_zipped_coords();
@@ -1085,7 +983,7 @@ SparseGlobalOrderReader::respect_copy_memory_budget(
         uint64_t idx = 0;
         for (; idx < max_cs_idx; idx++) {
           auto rt =
-              (ResultTileWithBitmap<uint8_t>*)result_cell_slabs[idx].tile_;
+              static_cast<GlobalOrderResultTile*>(result_cell_slabs[idx].tile_);
           auto id =
               std::pair<uint64_t, uint64_t>(rt->frag_idx(), rt->tile_idx());
           if (accounted_tiles.count(id) == 0) {
@@ -1170,9 +1068,8 @@ uint64_t SparseGlobalOrderReader::compute_var_size_offsets(
     while (query_buffer.original_buffer_var_size_ < new_var_buffer_size) {
       // Revert progress for this slab in read state, and pop it.
       auto& last_rcs = result_cell_slabs.back();
-      read_state_.frag_tile_idx_[last_rcs.tile_->frag_idx()] =
-          std::pair<uint64_t, uint64_t>(
-              last_rcs.tile_->tile_idx(), last_rcs.start_);
+      read_state_.frag_idx_[last_rcs.tile_->frag_idx()] =
+          FragIdx(last_rcs.tile_->tile_idx(), last_rcs.start_);
       result_cell_slabs.pop_back();
 
       // Update the new var buffer size.
@@ -1200,9 +1097,8 @@ uint64_t SparseGlobalOrderReader::compute_var_size_offsets(
     new_var_buffer_size = ((OffType*)query_buffer.buffer_)[total_cells];
 
     // Update the cell progress.
-    read_state_.frag_tile_idx_[last_rcs.tile_->frag_idx()] =
-        std::pair<uint64_t, uint64_t>(
-            last_rcs.tile_->tile_idx(), last_rcs.start_ + last_rcs.length_);
+    read_state_.frag_idx_[last_rcs.tile_->frag_idx()] =
+        FragIdx(last_rcs.tile_->tile_idx(), last_rcs.start_ + last_rcs.length_);
   }
 
   return new_var_buffer_size;
@@ -1364,8 +1260,7 @@ Status SparseGlobalOrderReader::process_slabs(
 }
 
 Status SparseGlobalOrderReader::remove_result_tile(
-    const unsigned frag_idx,
-    std::list<ResultTileWithBitmap<uint8_t>>::iterator rt) {
+    const unsigned frag_idx, TileListIt rt) {
   // Remove coord tile size from memory budget.
   auto tile_idx = rt->tile_idx();
   auto&& [st, tiles_sizes] = get_coord_tiles_size<uint8_t>(
@@ -1406,7 +1301,7 @@ Status SparseGlobalOrderReader::end_iteration() {
       storage_manager_->compute_tp(), 0, fragment_num, [&](uint64_t f) {
         while (!result_tiles_[f].empty() &&
                result_tiles_[f].front().tile_idx() !=
-                   read_state_.frag_tile_idx_[f].first) {
+                   read_state_.frag_idx_[f].tile_idx_) {
           RETURN_NOT_OK(remove_result_tile(f, result_tiles_[f].begin()));
         }
 
